from mishmash.orm import (Artist, Album, Library, VARIOUS_ARTISTS_NAME,
                          MAIN_LIB_ID, NULL_LIB_ID)
from .factories import (EpFactory, LibraryFactory, LpFactory,
                        DirectoryStructure)

def _isInitialDb(db):
    return db.query(Artist).one().name == VARIOUS_ARTISTS_NAME

'''
def test_library(tmpdir, database, mishmash_cmd):
    session = database.SessionMaker()
    assert _isInitialDb(session)

    lib = (LibraryFactory()
              .add(EpFactory())
              .add(LpFactory())
              .add(LpFactory())
              .add(EpFactory()))

    dir_struct = DirectoryStructure.PREFERRED
    for album in lib.albums():
        dir_struct.apply(*[t._mp3_file for t in album.tracks])
        
    #dir_struct.apply(*[t._mp3_file for t in [alb.tracks
    #                                            for alb in lib.albums()]],
    #                 root_dir=tmpdir)

    #mishmash_cmd(["sync", str(tmpdir)], db_url=database.url)
'''


def test_lpSync(tmpdir, database, mishmash_cmd):
    session = database.SessionMaker()
    lib_count = session.query(Library).count()
    assert lib_count == 2

    lp = LpFactory(temp_dir=str(tmpdir))

    dir_struct = DirectoryStructure.PREFERRED
    dir_struct.apply(*[t._mp3_file for t in lp.tracks], root_dir=tmpdir)

    mishmash_cmd(["info"], db_url=database.url)
    assert _isInitialDb(session)
    mishmash_cmd(["sync", str(tmpdir)], db_url=database.url)

    assert session.query(Library).count() == lib_count

    assert session.query(Artist).filter_by(lib_id=MAIN_LIB_ID).count() == 1
    assert (session.query(Artist).filter_by(lib_id=MAIN_LIB_ID,
                                            name=lp.artist).count() == 1 or
            session.query(Artist).filter_by(lib_id=MAIN_LIB_ID,
                                            name=lp.artist[0:30]).count() == 1
           )

    assert session.query(Album).filter_by(lib_id=MAIN_LIB_ID).count() == 1
    assert (session.query(Album).filter_by(lib_id=MAIN_LIB_ID,
                                           title=lp.title).count() == 1 or
            session.query(Artist).filter_by(lib_id=MAIN_LIB_ID,
                                            name=lp.artist[0:30]).count() == 1)

    lp = session.query(Album)\
                .filter((Library.id == MAIN_LIB_ID) &
                        ((Album.title == lp.title) |
                         (Album.title == lp.title[:30])))\
                .one()
    assert lp.type == "lp"

    # TODO: validate tracks


def test_epSync(tmpdir, database, mishmash_cmd):
    db = database.SessionMaker()
    lib_count = db.query(Library).count()
    assert lib_count == 2

    ep = EpFactory(artist="Mikal Cronin", title="Tide 7\"", num_tracks=2,
                   track_titles=("Tide", "You Gotta Have Someone"),
                   temp_dir=str(tmpdir))

    dir_struct = DirectoryStructure.PREFERRED
    dir_struct.apply(*[t._mp3_file for t in ep.tracks], root_dir=tmpdir)

    mishmash_cmd(["info"], db_url=database.url)
    assert _isInitialDb(db)
    mishmash_cmd(["sync", str(tmpdir)], db_url=database.url)

    assert db.query(Library).count() == lib_count

    assert db.query(Artist).filter_by(lib_id=MAIN_LIB_ID).count() == 1
    assert db.query(Artist).filter_by(lib_id=MAIN_LIB_ID,
                                      name="Mikal Cronin").count() == 1

    assert db.query(Album).filter_by(lib_id=MAIN_LIB_ID).count() == 1
    assert db.query(Album).filter_by(lib_id=MAIN_LIB_ID,
                                     title='Tide 7"').count() == 1

    ep = db.query(Album).filter_by(lib_id=MAIN_LIB_ID, title='Tide 7"').one()
    assert ep.type in ("ep", None)

    # TODO: validate tracks




import time
import subprocess

from multiprocessing import Process
from ..core import Command, CommandError


class MishMashProc(Process):
    def __init__(self, cmd, *args, config=None):
        self._cmd = cmd
        super().__init__(target=MishMashProc._entryPoint, args=[config, cmd, *args])

    @staticmethod
    def _entryPoint(config, cmd, *args):
        from ..__main__ import MishMash
        return MishMash(config_obj=config).run(args_list=[cmd, *args])

    def __str__(self):
        if self.exitcode is None:
            return f"`mishmash {self._cmd}` <running>"
        else:
            return f"`mishmash {self._cmd}` <stopped[{self.exitcode}]>"

    def start(self):
        super().start()
        return self

    def join(self, timeout=None, check=False) -> None:
        super().join(timeout)
        if check:
            if self.exitcode:
                raise CommandError(self)


class UnsonicProc:
    def __init__(self, config):
        self._config = config
        self._unsonic = None
        self.exitcode = None

    def start(self):
        self._unsonic = subprocess.Popen([
            "unsonic", "--config", self._config, "serve"
        ])
        return self

    def join(self, timeout=None, check=False) -> None:
        if self.exitcode is not None:
            return

        try:
            self._unsonic.wait(timeout=timeout)
        except subprocess.TimeoutExpired:
            self.exitcode = None
            return

        self.exitcode = self._unsonic.returncode
        if check and self.exitcode is not None and self.exitcode:
            raise CommandError(self)

    def kill(self):
        self._unsonic.kill()


@Command.register
class Server(Command):
    NAME = "server"
    HELP = "Sync, monitor, browse, etc."

    def _run(self):
        """main"""
        all_procs = []

        # `mishmash info`: tests the config and get banner and info.
        MishMashProc("info", config=self.args.config).start().join(check=True)

        sync = MishMashProc("sync", "--no-prompt", "--monitor", config=self.args.config) \
                if self.args.config.getboolean("server", "sync", fallback=True) else None
        web = MishMashProc("web", config=self.args.config) \
                if self.args.config.getboolean("server", "web", fallback=True) else None
        unsonic = self._createUnsonic()

        for p in (sync, web, unsonic):
            if p:
                all_procs.append(p)
                p.start()

        try:
            while True:
                for proc in all_procs:
                    proc.join(1, check=True)
                time.sleep(5)
        finally:
            for proc in all_procs:
                proc.kill()

    def _createUnsonic(self):
        if self.args.config.getboolean("server", "unsonic", fallback=False):
            cfg_file = self.args.config.get("unsonic", "config")
            return UnsonicProc(cfg_file)



from pathlib import Path
import eyed3.id3
from eyed3.core import LP_TYPE, EP_TYPE
from . import factories


def test_TagFactory():
    tag = factories.TagFactory.build()
    assert isinstance(tag, eyed3.id3.Tag)
    assert tag.title.startswith("Track title #")
    assert tag.artist == "Artist"
    assert tag.album == "Album"
    assert tag.album_artist == None
    assert tag.track_num == (None, None)

    tag2 = factories.TagFactory.build(title="Another title", track_num=2)
    assert tag2.title == "Another title"
    assert tag2.artist == tag.artist
    assert tag2.album == tag.album
    assert tag2.album_artist == None
    assert tag2.track_num == (2, None)


def test_Mp3AudioFileFactory(mp3audiofile):
    assert mp3audiofile.path is not None
    assert Path(mp3audiofile.path).exists()
    assert mp3audiofile.tag is not None
    assert mp3audiofile.tag.title is not None
    assert mp3audiofile.tag.artist is not None
    assert mp3audiofile.tag.album_artist is None


def test_AlbumFactory():
    album = factories.AlbumFactory()
    assert album.type == LP_TYPE
    assert factories.LP_SIZE[0] <= len(album.tracks) <= factories.LP_SIZE[1]

    title = album.title
    assert title
    artist = album.artist
    assert artist
    ogdate = album.original_release_date
    assert ogdate
    rdate = album.release_date

    tracks = album.tracks
    assert all([bool(t.artist.name == artist) for t in tracks])
    assert all([bool(t.album.title == title) for t in tracks])
    assert all([bool(t.album.original_release_date == ogdate) for t in tracks])
    assert all([(str(t.album.release_date), str(rdate)) for t in tracks])


def test_EpFactory():
    ep = factories.EpFactory()
    assert ep.type == EP_TYPE
    assert factories.EP_SIZE[0] <= len(ep.tracks) <= factories.EP_SIZE[1]


def test_LibraryFactory():
    lib = factories.LibraryFactory()
    assert lib.name
    assert lib.last_sync is None
    assert lib.id is None

    lib = factories.LibraryFactory(name="Music")
    assert lib.name == "Music"



from ..web import MISHMASH_WEB

if MISHMASH_WEB:
    import tempfile
    from pathlib import Path
    from ..core import Command
    from pyramid.scripts.pserve import PServeCommand

    @Command.register
    class Web(Command):
        NAME = "web"
        HELP = "MishMash web interface."

        def _initArgParser(self, parser):
            parser.add_argument("-p", "--port", type=int, default=None)

        def _run(self):
            if self.args.port:
                self.config["server:main"]["port"] = str(self.args.port)

            # pserve wants a file to open, so use the composed config.

            with tempfile.NamedTemporaryFile(mode="w", suffix=".ini", delete=False) as config_file:
                self.config.write(config_file)
                config_file.flush()
                pserve = PServeCommand(["mishmash", config_file.name])
                try:
                    return pserve.run()
                finally:
                    tmp_cfg = Path(config_file.name)
                    if tmp_cfg.exists():
                        # Must clean only only once and multiple web workers are spawned
                        tmp_cfg.unlink()



from . import info                                                # noqa
from . import sync                                                # noqa
from . import mgmt                                                # noqa
from . import web                                                 # noqa
from . import server                                              # noqa



import random
import tempfile
import subprocess
from enum import Enum
from pathlib import Path

import faker
import factory
import factory.fuzzy
import eyed3.id3
import eyed3.mp3
from eyed3.core import Date, AudioFile, LP_TYPE, EP_TYPE
from eyed3.id3 import (ID3_V1_0, ID3_V1_1,
                       ID3_V2_2, ID3_V2_3, ID3_V2_4)
from faker.providers import BaseProvider
from mishmash.core import EP_MAX_SIZE_HINT
from mishmash.orm import Artist, Album, Track, Library

EP_SIZE = (2, EP_MAX_SIZE_HINT)
LP_SIZE = (EP_MAX_SIZE_HINT + 1, 24)


class DirectoryStructure(Enum):
    NONE = 0
    PREFERRED = 1

    def structuredPath(self, file):
        assert isinstance(file, AudioFile)
        tag = file.tag
        assert tag.file_info.name == file.path

        if self == DirectoryStructure.NONE:
            return file.path
        else:
            assert self == DirectoryStructure.PREFERRED
            return Path().joinpath(
                f"{tag.artist}",
                f"{tag.original_release_date} - {tag.album}",
                f"{tag.artist} - {tag.track_num[0]} - {tag.title}")

    def apply(self, *files, root_dir="./"):
        root_dir = Path(root_dir)
        for file in files:
            curr_path = Path(file.path)
            new_path = root_dir / self.structuredPath(file)

            if not new_path.parent.exists():
                new_path.parent.mkdir(parents=True)

            if curr_path.name != new_path.name:
                file.rename(str(new_path))


class TagFactory(factory.Factory):
    class Params:
        id3_version = factory.Faker("id3_version")

    class Meta:
        model = eyed3.id3.Tag

    title = factory.Sequence(lambda n: "Track title #{:d}".format(n))
    artist = "Artist"
    album = "Album"
    album_artist = None
    track_num = None

    @classmethod
    def create_batch(cls, size, version, track_titles=None, **kwargs):
        tags = super().create_batch(size, **kwargs)
        for i, t in enumerate(tags, 1):
            t.version = version
            if track_titles and i <= len(track_titles):
                t.title = track_titles[i - 1]
            t.track_num = (i, size)
        return tags


class Mp3AudioFileFactory(factory.Factory):
    _TEMP_D = None
    _SAMPLE_MP3 = subprocess.run("dd if=/dev/zero bs=1MB count=10 | lame -r -",
                                 shell=True, check=True,
                                 stdout=subprocess.PIPE).stdout

    class Meta:
        model = eyed3.mp3.Mp3AudioFile

    class Params:
        id3_version = factory.Faker("id3_version")
        temp_dir = None

    @factory.lazy_attribute
    def path(obj):
        tmp_dir = str(obj.temp_dir) if obj.temp_dir \
                                    else Mp3AudioFileFactory.getTempDir()
        mp3 = tempfile.NamedTemporaryFile(dir=str(tmp_dir),
                                          suffix=".mp3", delete=False)
        mp3.write(Mp3AudioFileFactory._SAMPLE_MP3)
        mp3.close()
        return mp3.name

    @factory.post_generation
    def tag(obj, create, extracted, **kwargs):
        if create:
            obj.tag = extracted
            obj.tag.save(version=obj.tag.version)

    @classmethod
    def getTempDir(cls):
        if cls._TEMP_D is None:
            cls._TEMP_D = tempfile.TemporaryDirectory()
        return Path(cls._TEMP_D.name)


class ArtistFactory(factory.Factory):
    class Meta:
        model = Artist


class AlbumFactory(factory.Factory):
    class Params:
        num_tracks = factory.fuzzy.FuzzyInteger(*LP_SIZE)
        dir_structure = DirectoryStructure.PREFERRED
        id3_version = factory.Faker("id3_version")
        track_titles = []
        temp_dir = None

    class Meta:
        model = Album

    type = LP_TYPE
    title = factory.Faker("id3_title")
    artist = factory.Faker("id3_artist")
    original_release_date = factory.Faker("id3_date")
    # TODO: Must me >= original_release_date
    release_date = original_release_date

    @factory.lazy_attribute
    def tracks(obj):
        tags = TagFactory.create_batch(obj.num_tracks, obj.id3_version,
                                       obj.track_titles,
                                       artist=obj.artist,
                                       album=obj.title)
        assert len(tags) == obj.num_tracks
        assert len(set([t.track_num[0] for t in tags])) == obj.num_tracks
        assert all([bool(t.album == obj.title) for t in tags])
        assert all([bool(t.artist == obj.artist) for t in tags])

        tracks = []
        for t in tags:
            t.album_type = obj.type
            t.original_release_date = obj.original_release_date
            if obj.id3_version[0] != 1:
                # TODO: allow for different, but greater than, values
                t.release_date = obj.original_release_date

            mp3 = Mp3AudioFileFactory(tag=t, temp_dir=obj.temp_dir)
            assert hasattr(t, "file_info") and t.file_info.name,\
                        "Tag is not attached to a file"

            track = Track().update(mp3)
            track._mp3_file = mp3
            track.artist = ArtistFactory(name=obj.artist)
            # Note, cannot call AlbumFactory here as calling this function is
            # required to do so.
            track.album = Album(title=obj.title, type=obj.type,
                                artist=obj.artist,
                                original_release_date=obj.original_release_date,
                                release_date=obj.release_date)
            tracks.append(track)

        return tracks


class EpFactory(AlbumFactory):
    class Params:
        num_tracks = factory.fuzzy.FuzzyInteger(*EP_SIZE)
    type = EP_TYPE


class LpFactory(AlbumFactory):
    class Params:
        num_tracks = factory.fuzzy.FuzzyInteger(*LP_SIZE)
    type = LP_TYPE


class LibraryFactory(factory.Factory):

    class Meta:
        model = Library

    name = factory.Faker("id3_title")
    last_sync = None

    '''
    def defaultDir(self, tmpdir):
        self._tmdir = tmpdir

    def addEp(self, *args, **kwargs):
        # TODO
        # TODO: use self._tmpdir if set
        return self
    '''


class Id3Provider(BaseProvider):
 _fake = faker.Faker()
 V1_VERSION_WEIGHTS = [(ID3_V1_0, 5), (ID3_V1_1, 15)]
 V2_VERSION_WEIGHTS = [(ID3_V2_2, 5), (ID3_V2_3, 60), (ID3_V2_4, 40)]
 VERSION_WEIGHTS = V1_VERSION_WEIGHTS + V2_VERSION_WEIGHTS

 def id3_version(self):
     return random.choices(
         [v for v, _ in self.VERSION_WEIGHTS],
         cum_weights=[w for _, w in self.VERSION_WEIGHTS])[0]

 def id3_title(self):
     return " ".join(self._fake.words(random.randint(1, 5)))

 def id3_artist(self):
     return " ".join(self._fake.words(random.randint(1, 3)))

 def id3_date(self):
     d = self._fake.date_object()
     full_date = random.choices(["y", "n"], cum_weights=[15, 85]) == "y"
     return Date(year=d.year,
                 month=d.month if full_date else None,
                 day=d.day if full_date else None)

 def id3_tag(self, **kwargs):
     return TagFactory(**kwargs)


factory.Faker.add_provider(Id3Provider)



try:
    import pyramid                                                  # noqa: F401
    MISHMASH_WEB = True
except ImportError:
    MISHMASH_WEB = False
else:
    from pyramid.config import Configurator
    from zope.sqlalchemy import register as zope_tranaction_register

    from .. import database
    from ..config import Config

    def _configure(settings, DBSession):
        config = Configurator(settings=settings)

        config.include('pyramid_chameleon')
        config.include('pyramid_layout')

        def _DBSession(request):
            return DBSession()
        config.add_request_method(_DBSession, name="DBSession", reify=True)

        config.add_static_view('static', 'mishmash.web:static',
                               cache_max_age=3600)

        config.add_route('all_artists', '/artists')
        config.add_route('all_albums', '/albums')
        config.add_route('artist', '/artist/{id:\d+}')  # noqa: W605
        config.add_route('images.covers', '/images/covers/{id:\d+|default}')  # noqa: W605
        config.add_route('images.artist', '/images/artist/{id:\d+|default}')  # noqa: W605
        config.add_route('home', '/')
        config.add_route('search', '/search')
        config.add_route('new_music', '/new')
        config.add_route('album', '/album/{id:\d+}')  # noqa: W605

        config.scan(".panels")
        config.scan(".layouts")
        config.scan(".views")

        return config

    def main(global_config, **main_settings):
        app_config = Config(global_config["__file__"])
        app_config.read()
        mm_settings = app_config["mishmash"]

        engine_args = dict(database.DEFAULT_ENGINE_ARGS)
        pfix, plen = "sqlalchemy.", len("sqlalchemy.")
        # Strip prefix and remove url value
        sql_ini_args = {
                name[plen:]: mm_settings[name]
                for name in mm_settings
                if name.startswith(pfix) and not name.endswith(".url")
        }
        engine_args.update(sql_ini_args)

        (engine,
         SessionMaker,
         connection) = database.init(app_config.db_url, engine_args=engine_args, scoped=True,
                                     trans_mgr=zope_tranaction_register)

        pyra_config = _configure(main_settings, SessionMaker)
        return pyra_config.make_wsgi_app()



import uuid
import shutil
import tempfile
import pytest
from pathlib import Path
from collections import namedtuple
from tempfile import NamedTemporaryFile

import mishmash.orm
import mishmash.database
from mishmash.__main__ import MishMash
from .factories import Mp3AudioFileFactory, TagFactory, LibraryFactory

TestDatabase = namedtuple("TestDatabase", ["url", "engine", "SessionMaker"])


@pytest.fixture(scope="function",
                params=["sqlite", "postgresql"])
def database(request, pg_server):
    db_file = None
    if request.param == "sqlite":
        db_file = NamedTemporaryFile(suffix=".sqlite", delete=False)
        db_file.close()
        db_url = f"sqlite:///{db_file.name}"
    elif request.param == "postgresql":
        db_url = "postgresql://{user}:{password}@{host}:{port}/{database}"\
                 .format(**pg_server["params"])
    else:
        assert not("unhandled db: " + request.param)
        return

    engine, SessionMaker, connection = mishmash.database.init(db_url)

    # Outermost transaction that is always rolled back
    trans = connection.begin()
    yield TestDatabase(url=db_url, engine=engine, SessionMaker=SessionMaker)

    # ... teardown
    trans.rollback()
    connection.close()

    if request.param == "sqlite":
        assert db_file
        Path(db_file.name).unlink()
    else:
        mishmash.database.dropAll(db_url)


@pytest.fixture
def session(database):
    session = database.SessionMaker()

    yield session

    # ... teardown
    session.rollback()
    session.close()


@pytest.fixture(scope="session", autouse=True)
def mishmash_tempdir():
    global TEMP_DIR

    temp_d = tempfile.TemporaryDirectory()
    TEMP_DIR = temp_d

    yield Path(temp_d.name)

    if Path(temp_d.name).exists():
        temp_d.cleanup()


@pytest.fixture(scope="function")
def mp3audiofile(mishmash_tempdir):
    mp3_file = Mp3AudioFileFactory(tag=TagFactory())

    yield mp3_file

    Path(mp3_file.path).unlink()


def _tempCopy(src, dest_dir):
    testfile = Path(str(dest_dir)) / "{}.mp3".format(uuid.uuid4())
    shutil.copyfile(str(src), str(testfile))
    return testfile


@pytest.fixture(scope="function")
def db_library(session):
    lib = LibraryFactory()
    session.add(lib)
    session.commit()

    yield lib

    # ... teardown
    # the session is getting popped, so do nothing


@pytest.fixture(scope="session")
def mishmash_cmd():
    def func(args, expected_retval=0, db_url=None):
        if database:
            args = ["--database", db_url] + args

        app = MishMash()
        try:
            retval = app._run(args)
        except SystemExit as exit:
            retval = exit.code
        assert retval == expected_retval

    return func



import gettext
from pyramid_layout.layout import layout_config

_ = gettext.gettext


@layout_config(
    name='main-layout',
    template='mishmash.web:templates/layouts/main-layout.pt'
)
class AppLayout:

    def __init__(self, context, request):
        self.context = context
        self.request = request
        self.home_url = request.application_url
        self.headings = []

    @property
    def page_title(self):
        return _("MishMash music!")

    def add_heading(self, name, *args, **kw):
        self.headings.append((name, args, kw))



import sys
from nicfit.console.ansi import Fg, Style
from pyfiglet import figlet_format
from sqlalchemy.exc import ProgrammingError, OperationalError
from .. import version
from ..core import Command
from ..util import safeDbUrl
from ..orm import Track, Artist, Album, Meta, Tag, Library

"""
TODO:
    - command line arg for selecting libary
"""


class DisplayList:
    def __init__(self):
        self._rows = []

    def add(self, key, val):
        self._rows.append(tuple((key, val)))

    def print(self, _format, clear=False, **kwargs):
        k_width = max([len(k) for k, v in self._rows if k])
        for k, v in self._rows:
            if k:
                print(_format.format(k=k.ljust(k_width), v=v, **kwargs))

        if clear:
            self.clear()

    def clear(self):
        self._rows.clear()


@Command.register
class Info(Command):
    NAME = "info"
    HELP = "Show information about the database and configuration."
    _library_arg_nargs = "*"

    def _initArgParser(self, parser):
        super()._initArgParser(parser)
        parser.add_argument("--artists", dest="show_artists",
                            action="store_true",
                            help="List all artists, per library.")

    def lib_query(self, OrgType, lib):
        if isinstance(lib, int):
            lid = lib
        else:
            lid = lib.id
        return self.db_session.query(OrgType).filter_by(lib_id=lid)

    def _displayMetaInfo(self):
        display_list = DisplayList()

        def mkkey(k):
            return Style.bright(Fg.blue(str(k)))

        def mkval(v):
            return str(v)

        display_list.add(mkkey("Version"), mkval(version))
        display_list.add(mkkey("Database URL"),
                         mkval(safeDbUrl(self.config.db_url)))
        try:
            meta = self.db_session.query(Meta).one()
        except (ProgrammingError, OperationalError) as ex:
            print("\nError querying metadata. Database may not be "
                  "initialized: %s" % str(ex), file=sys.stderr)
            return 1

        display_list.add(mkkey("Database version"), mkval(meta.version))
        display_list.add(mkkey("Last sync"), mkval(meta.last_sync or "Never"))
        display_list.add(mkkey("Configuration files "),
                         mkval(", ".join(self.args.config.input_filenames)))
        display_list.print("{k} {delim} {v}", delim=Style.bright(":"))

    def _displayLibraryInfo(self, lib):
        def mkkey(k):
            return Style.bright(str(k))

        display_list = DisplayList()
        for name, orm_type in [("tracks", Track), ("artists", Artist),
                               ("albums", Album), ("tags", Tag),
                              ]:
            count = self.lib_query(orm_type, lib).count()
            display_list.add(mkkey(count), name)

        display_list.print("{k} music {v}", clear=True)

    def _displayArtists(self, lib):
        for a in self.lib_query(Artist, lib).order_by(Artist.sort_name).all():
            print(a.name)

    def _run(self):
        logo = figlet_format("``MishMash``", font="graffiti")
        print(Fg.green(logo, Style.BRIGHT))

        self._displayMetaInfo()

        for lib in Library.iterall(self.db_session, names=self.args.libs):
            if self.args.show_artists:
                print(Fg.green(f"\n=== {lib.name} library artists ==="))
                self._displayArtists(lib)
            else:
                print(Fg.green(f"\n=== {lib.name} library ==="))
                self._displayLibraryInfo(lib)



import os
import random
from hashlib import sha1
from gettext import gettext as _

from pyramid.response import Response
from pyramid.view import view_config
from pyramid.httpexceptions import HTTPNotFound, HTTPNotModified

from sqlalchemy import desc

from eyed3.utils import formatTime
from eyed3.core import (LP_TYPE, EP_TYPE, COMP_TYPE, VARIOUS_TYPE, LIVE_TYPE,
                        DEMO_TYPE, SINGLE_TYPE, ALBUM_TYPE_IDS)

from ..__about__ import __project_name__ as PROJECT_NAME
from ..__about__ import __version__ as VERSION

from .. import database
from .. import util
from ..orm import Artist, Album, Image, VARIOUS_ARTISTS_NAME, VARIOUS_ARTISTS_ID

_vars = {"project_name": PROJECT_NAME,
         "project_version": VERSION,
        }


class ResponseDict(dict):
    def __init__(self, *args, **kwargs):
        super(ResponseDict, self).__init__(*args, **kwargs)
        self.update(_vars)


# Not in eyeD3
ALL_TYPE = "All"
TYPE_DISPLAY_NAMES = {ALL_TYPE: _(ALL_TYPE),
                      LP_TYPE: _("LPs"),
                      EP_TYPE: _("EPs"),
                      COMP_TYPE: _("Compilations"),
                      VARIOUS_TYPE: VARIOUS_ARTISTS_NAME,
                      LIVE_TYPE: _("Live"),
                      DEMO_TYPE: _("Demos"),
                      SINGLE_TYPE: _("Singles"),
                     }
TYPE_DISPLAY_NAMES[ALL_TYPE] = _(ALL_TYPE)


@view_config(route_name="home", renderer="templates/home.pt",
             layout="main-layout")
def home_view(request):
    return ResponseDict()


@view_config(route_name="all_artists", renderer="templates/artists.pt",
             layout="main-layout")
def allArtistsView(request):
    NUM_COLS = 2
    NUMBER = "#"
    OTHER = "Other"
    buckets = {}   # buckets['A'] = count, etc.
    artist_dict = {}

    def _whichBucket(name):
        first_l = name[0].upper()
        if not first_l.isalpha():
            first_l = NUMBER if first_l.isnumeric() else OTHER
        if first_l not in buckets:
            buckets[first_l] = 0
        buckets[first_l] += 1
        return first_l

    count = 0
    active_types = _getActiveAlbumTypes(request.params)
    artist_types = {}
    session = request.DBSession
    for artist in session.query(Artist)\
                         .order_by(Artist.sort_name).all():

        types = set([alb.type for alb in artist.albums])
        for t in types:
            if t not in artist_types:
                artist_types[t] = {
                    "active": t in active_types,
                    "display_name": TYPE_DISPLAY_NAMES[t],
                }

        if not active_types.intersection(types):
            continue

        count += 1

        bucket = _whichBucket(artist.sort_name)
        if bucket not in artist_dict:
            artist_dict[bucket] = []

        if artist_dict[bucket]:
            # Adding show_origin member to the instance here. True for
            # dup artists, false otherwise.
            if artist_dict[bucket][-1].name == artist.name:
                artist_dict[bucket][-1].show_origin = True
                artist.show_origin = True
            else:
                artist.show_origin = False
        else:
            artist.show_origin = False

        artist_dict[bucket].append(artist)

    keys = list(buckets.keys())
    keys.sort()
    if OTHER in keys:
        keys.remove(OTHER)
        keys.append(OTHER)

    # Build column based on keys and counts
    col_cnt, col_curr = 0, 0
    key_cols = []   # key_cols[0] = [k1, k2, ...], key_cols[1] = [k10, ...]
    key_cols.append([])
    num_per_col = round(count / NUM_COLS)
    for k in keys:
        col_cnt += buckets[k]
        if col_cnt >= num_per_col:
            key_cols.append([])
            col_curr += 1
            col_cnt = 0

        key_cols[col_curr].append(k)

    return ResponseDict(artist_keys=keys, artist_dict=artist_dict, artist_types=artist_types,
                        artist_count=count, key_columns=key_cols)


@view_config(route_name="artist", renderer="templates/artist.pt",
             layout="main-layout")
def artistView(request):
    artist_id = int(request.matchdict["id"])
    session = request.DBSession

    artist = session.query(Artist).filter_by(id=artist_id).first()
    if not artist:
        raise HTTPNotFound()

    def _filterByType(typ, artist, albums):
        if typ == SINGLE_TYPE:
            return artist.getTrackSingles()
        elif typ == ALL_TYPE:
            return albums
        else:
            return artist.getAlbumsByType(typ)

    albums = list(artist.albums)
    all_tabs = [ALL_TYPE] + ALBUM_TYPE_IDS

    active_albums = []
    active_singles = []
    active_tab = request.GET.get("album_tab", None)
    if not active_tab:
        # No album type was requested, try to pick a smart one.
        for active_tab in all_tabs:
            active_albums = _filterByType(active_tab, artist, albums)
            if active_albums:
                break
    else:
        active_albums = _filterByType(active_tab, artist, albums)

    if active_tab == SINGLE_TYPE:
        active_singles = active_albums
        active_albums = []

    if active_albums:
        active_albums = util.sortByDate(active_albums,
                                        active_tab == LIVE_TYPE)
    else:
        # Unlike tags, the orm.Track does not have dates so not sorting :/
        # XXX active_singles = util.sortByDate(active_singles)
        pass

    for a in active_albums:
        covers = [img for img in a.images
                        if img.type == Image.FRONT_COVER_TYPE]
        a.cover = random.choice(covers) if covers else None

    tabs = {}
    if artist.id != VARIOUS_ARTISTS_ID:
        for name in all_tabs:
            t = dict(name=name, display_name=TYPE_DISPLAY_NAMES[name], active=active_tab == name,
                     has_items=bool(len(_filterByType(name, artist, albums))))
            if name in tabs:
                raise ValueError("Duplicate album type found: " + name)
            tabs[name] = t

        if len([t for t in tabs.values() if t["has_items"]]) == 2:
            # If all we have is "All" and another type, lose "All"
            tabs[ALL_TYPE]["has_items"] = False

    return ResponseDict(artist=artist,
                        active_tab=active_tab,
                        active_albums=active_albums,
                        active_singles=active_singles,
                        tabs=tabs,
                       )


@view_config(route_name="images.covers")
def covers(request):
    return _imageView(request, default_resp=Response(content_type="image/png",
                                                     body=DEFAULT_COVER_DATA))


@view_config(route_name="images.artist")
def artist_images(request):
    return _imageView(request)


with open(os.path.join(os.path.dirname(__file__),
                       "static",
                       "record150.png"), "rb") as _cover_fp:
    DEFAULT_COVER_DATA = _cover_fp.read()
del _cover_fp


@view_config(route_name="new_music", renderer="templates/new_music.pt",
             layout="main-layout")
def newMusicView(request):
    from math import ceil
    session = request.DBSession
    albums = session.query(Album).order_by(desc("date_added")).limit(25).all()
    return ResponseDict(albums=albums, ceil=ceil)


@view_config(route_name="album", renderer="templates/album.pt",
             layout="main-layout")
def albumView(request):
    album_id = int(request.matchdict["id"])
    session = request.DBSession

    album = session.query(Album).filter_by(id=album_id).first()
    if not album:
        raise HTTPNotFound()
    return ResponseDict(album=album,
                        formatTime=formatTime,
                       )


@view_config(route_name="search", renderer="templates/search_results.pt",
             layout="main-layout")
def searchView(request):
    query = request.POST["q"]
    results = database.search(request.DBSession, query)
    return ResponseDict(**results)


@view_config(route_name="all_albums", renderer="templates/albums.pt",
             layout="main-layout")
def allAlbumsView(request):
    album_dict = {}
    album_types = {}

    if ("filter-form" in request.params and request.params.get("filter-form") == "true"
            and not request.params.getall("type")):
        # Form submit with no filters checks, this is not the default /albums
        active_types = set()
    else:
        active_types = _getActiveAlbumTypes(request.params)

    session = request.DBSession
    for album in session.query(Album)\
                        .order_by(Album.original_release_date).all():

        if album.type not in album_types:
            album_types[album.type] = {
                "active": album.type in active_types,
                "display_name": TYPE_DISPLAY_NAMES[album.type],
            }

        if not album_types[album.type]["active"]:
            continue

        d = album.getBestDate()
        bucket = d.year // 10 * 10 if d else 0
        if bucket not in album_dict:
            album_dict[bucket] = []

        album_dict[bucket].append(album)

    buckets = list(album_dict.keys())
    buckets.sort()

    return ResponseDict(album_decades=buckets, album_dict=album_dict, album_types=album_types)


def _imageView(request, default_resp=None):
    iid = request.matchdict["id"]

    if iid == "default" and default_resp:
        resp = default_resp
    else:
        session = request.DBSession
        image = session.query(Image).filter(Image.id == int(iid)).first()
        if not image:
            raise HTTPNotFound()
        resp = Response(content_type=image.mime_type, body=image.data)

    hash = sha1()
    hash.update(resp.body)
    etag = hash.hexdigest()
    if "If-None-Match" in request.headers and request.headers["If-None-Match"] == etag:
        raise HTTPNotModified()

    resp.headers["Cache-Control"] = "max-age=3600"
    resp.headers["ETag"] = etag

    return resp


def _getActiveAlbumTypes(params):
    inc_types = set([p for p in params.getall("type") if p[0] != '!'])
    exc_types = set([p[1:] for p in params.getall("type") if p[0] == '!'])
    active_types = set(ALBUM_TYPE_IDS).difference(exc_types)
    if inc_types:
        active_types = active_types.intersection(inc_types)
    return active_types



# -*- coding: utf-8 -*-
import pytest
from mishmash.config import DEFAULT_CONFIG, Config


def test_Config_nofile():
    c = Config(None)
    assert c.filename is None
    with pytest.raises(Exception):
        assert c.db_url is None
    with pytest.raises(Exception):
        assert c.music_libs is None



"""Album unique constraints

Revision ID: 2008219c2861
Revises: 74df8fa5a35f
Create Date: 2019-02-09 20:58:43.822325

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '2008219c2861'
down_revision = '74df8fa5a35f'
branch_labels = None
depends_on = None


def upgrade():
    with op.batch_alter_table("tracks") as batch_op:
        batch_op.alter_column('metadata_format', existing_type=sa.VARCHAR(length=7),
                              nullable=False)


def downgrade():
    with op.batch_alter_table("tracks") as batch_op:
        batch_op.alter_column('metadata_format', existing_type=sa.VARCHAR(length=7),
                              nullable=True)



from .core import *   # noqa: F403
from . import user

# Core orm
TYPES = [Meta, Library, Tag, Artist, Album, Track, Image]  # noqa: F405
TAGS = [artist_tags, album_tags, track_tags, artist_images, album_images]  # noqa: F405
IMAGE_TABLES = [artist_images, album_images]  # noqa: F405
ENUMS = [Image._types_enum, Album._types_enum]  # noqa: F405

# User orm
TYPES += [user.User]

# All the table instances. Order matters (esp. for postgresql). The tables are created in
# normal order, and dropped in reverse order.
TABLES = [T.__table__ for T in TYPES] + TAGS + IMAGE_TABLES



import random
from pyramid_layout.panel import panel_config
from ..__about__ import __version__, __years__, __project_name__
from .. import orm


@panel_config(name='navbar',
              renderer='mishmash.web:templates/panels/navbar.pt')
def navbar(context, request):
    # XXX: has not been needed for a while, leaving as a panel in the case it is again.
    return {}


@panel_config(name='footer')
def footer(context, request):
    return f"""
<footer class="text-muted">
  <div class="container">
    <p class="float-right">
      <a href="#">Back to top</a>
    </p>
    <p><br/></p>
    <p align='right'>{__project_name__} {__version__} &copy; {__years__}</p>
  </div>
</footer>
"""


@panel_config(name='album_cover')
def album_cover(context, request, album, size=None, link=False):
    front_covers = [img for img in album.images
                        if img.type == orm.Image.FRONT_COVER_TYPE]
    cover_id = random.choice(front_covers).id if front_covers else "default"
    cover_url = request.route_url("images.covers", id=cover_id)
    width = str(size or "100%")
    height = str(size or "100%")

    panel = (
        "<img class='shadow' width='%s' height='%s' src='%s' title='%s'/>" %
        (width, height, cover_url, "%s - %s" % (album.artist.name, album.title))
    )

    if link:
        panel = "<a href='%s'>%s</a>" % \
                (request.route_url('album', id=album.id), panel)

    return panel


@panel_config(name='artist_image')
def artist_image(context, request, artist, scale_percent=None, link=False):
    imgs = [img for img in artist.images]
    if not imgs:
        return ""

    width = str(scale_percent or "100%")
    height = str(scale_percent or "100%")
    img_url = request.route_url("images.artist", id=random.choice(imgs).id)
    panel = f"<img class='shadow' width='{width}' height='{height}' "\
            f"src='{img_url}' title='{artist.name}'/>"
    if link:
        panel = f"<a href='{request.route_url('artist', id=artist.id)}'>{panel}</a>"

    return panel



import os
import sys
import traceback

from sqlalchemy import exc as sql_exceptions
from nicfit.console import ansi
from nicfit import Application, ConfigOpts

from eyed3.utils.prompt import PromptExit

from .config import DEFAULT_CONFIG, CONFIG_ENV_VAR, Config, MAIN_SECT, SA_KEY
from . import log
from . core import Command, CommandError
from .commands import *                                                   # noqa

_FORK_METHOD_SET = False


def _pErr(msg):
    print(ansi.Fg.red(str(msg) + ":"))
    tb = sys.exc_info()
    traceback.print_exception(tb[0], tb[1], tb[2])


def main(args):
    import multiprocessing
    global _FORK_METHOD_SET

    if not _FORK_METHOD_SET:
        try:
            multiprocessing.set_start_method("fork")
            _FORK_METHOD_SET = True
        except RuntimeError as ex:
            log.warning("multiprocessing.set_start_method: " + str(ex))

    if not hasattr(args, "command_func") or not args.command_func:
        # No command was given.
        args.app.arg_parser.print_help()
        return 1

    # In the case fileConfig undid the command line, which has precedence.
    args.applyLoggingOpts(args.log_levels, args.log_files)

    if args.db_url:
        args.config.set(MAIN_SECT, SA_KEY, args.db_url)
        # Don't want commands and such to use this, so reset.
        args.db_url = None
    elif "MISHMASH_DBURL" in os.environ:
        log.verbose("Using environment MISHMASH_DBURL over configuration: {}"
                    .format(os.environ["MISHMASH_DBURL"]))
        args.config.set(MAIN_SECT, SA_KEY, os.environ["MISHMASH_DBURL"])

    # Run command
    try:
        retval = args.command_func(args) or 0
    except (KeyboardInterrupt, PromptExit):
        # PromptExit raised when CTRL+D during prompt, or prompts disabled
        retval = 0
    except (sql_exceptions.ArgumentError,):
        _pErr("Database error")
        retval = 1
    except (sql_exceptions.OperationalError,) as db_err:
        print(str(db_err), file=sys.stderr)
        retval = 1
    except CommandError as cmd_err:
        print(str(cmd_err), file=sys.stderr)
        retval = cmd_err.exit_status
    except Exception as ex:
        log.exception(ex)
        _pErr("General error")
        retval = 2

    return retval


class MishMash(Application):
    def __init__(self, progname="mishmash", ConfigClass=None, config_obj=None):
        from . import version

        if ConfigClass and config_obj:
            raise ValueError("ConfigClass and config_obj are not compatible together.")

        config_opts = ConfigOpts(required=False,
                                 default_config=DEFAULT_CONFIG,
                                 default_config_opt="--default-config",
                                 ConfigClass=ConfigClass or Config,
                                 config_env_var=CONFIG_ENV_VAR,
                                 init_logging_fileConfig=True)
        super().__init__(main, name=progname, version=version,
                         config_opts=config_opts, pdb_opt=False,
                         gettext_domain="MishMash")

        ansi.init()

        desc = "Database command line options (or config) are required by most sub commands."
        subs = self.arg_parser.add_subparsers(title="Commands", dest="command",
                                              description=desc, required=False)
        Command.loadCommandMap(subparsers=subs)
        self.arg_subparsers = subs
        self._user_config = config_obj

    def _main(self, args):
        if self._user_config:
            args.config = self._user_config

        return super()._main(args)

    def _addArguments(self, parser):
        group = parser.add_argument_group(title="Settings and options")
        group.add_argument("-D", "--database", dest="db_url", metavar="url", default=None,
                           help="Database URL. This will override the URL from the config "
                                "file be it the default of one passed with -c/--config.")


app = MishMash()
if __name__ == "__main__":
    app.run()                                                 # pragma: no cover



import datetime
import pytest
from sqlalchemy.exc import IntegrityError, DataError
import mishmash
from mishmash.orm import (Meta, Library, Artist, Album, Track, Image,
                          NULL_LIB_ID, NULL_LIB_NAME,
                          MAIN_LIB_ID, MAIN_LIB_NAME)


def test_meta_table(session):
    metadata = session.query(Meta).one()
    assert metadata.version == mishmash.version
    assert metadata.last_sync is None


def test_meta_table_write(session):
    metadata = session.query(Meta).one()
    metadata.version = "1.2.3"
    t = datetime.datetime.utcnow()
    metadata.last_sync = t
    session.add(metadata)
    session.commit()

    metadata = session.query(Meta).one()
    assert metadata.version == "1.2.3"
    assert metadata.last_sync == t


def test_libraries_table(session):
    # Default rows, for __null__ and Music
    default_rows = session.query(Library).all()
    assert len(default_rows) == 2
    assert default_rows[0].id == NULL_LIB_ID
    assert default_rows[0].name == NULL_LIB_NAME
    assert default_rows[0].last_sync is None
    assert default_rows[1].id == MAIN_LIB_ID
    assert default_rows[1].name == MAIN_LIB_NAME
    assert default_rows[1].last_sync is None


def test_ArtistNoLib(session):
    session.add(Artist(name="Pantera"))
    with pytest.raises(IntegrityError):
        session.commit()

def test_Artist(session, db_library):
    lid = db_library.id
    session.add(Artist(name="The Roots", lib_id=lid))
    session.commit()
    assert session.query(Artist).filter_by(name="The Roots", lib_id=lid).one()


def test_Track(session, db_library, mp3audiofile):
    lid = db_library.id
    artist = Artist(name=mp3audiofile.tag.artist, lib_id=lid)
    session.add(artist)
    session.flush()
    session.add(Track(audio_file=mp3audiofile, lib_id=lid, artist_id=artist.id))
    session.commit()
    assert session.query(Track).filter_by(title=mp3audiofile.tag.title,
                                          lib_id=lid).one()


def test_MetaTooBig(session, db_library, request):
    lid = db_library.id
    v = "!" * (Meta.VERSION_LIMIT + 1)
    session.add(Meta(version=v))
    if "sqlite" in request.keywords:
        session.commit()
        assert session.query(Meta).filter_by(version=v).one()
    else:
        with pytest.raises(DataError):
            session.commit()


def test_LibraryTooBig(session, db_library, request):
    lid = db_library.id
    v = "!" * (Library.NAME_LIMIT + 1)
    session.add(Library(name=v))
    if "sqlite" in request.keywords:
        session.commit()
        assert session.query(Library).filter_by(name=v).one()
    else:
        with pytest.raises(DataError):
            session.commit()


def test_ColumnTooBigTruncates(session, db_library, request, mp3audiofile):
    artist_name = "*" * (Artist.NAME_LIMIT + 1)
    city = "c" * (Artist.CITY_LIMIT + 1)
    state = "s" * (Artist.STATE_LIMIT + 1)
    artist = Artist(name=artist_name, origin_city=city, origin_state=state,
                    origin_country="United States", lib_id=db_library.id)
    session.add(artist)
    session.commit()
    artist = session.query(Artist)\
                    .filter_by(name=artist_name[:Artist.NAME_LIMIT],
                               origin_city=city[:Artist.CITY_LIMIT],
                               origin_state=state[:Artist.STATE_LIMIT],
                               origin_country="USA").one()
    assert artist

    album_name = "*" * (Album.TITLE_LIMIT + 1)
    album = Album(title=album_name, artist_id=artist.id, lib_id=db_library.id)
    session.add(album)
    session.commit()
    album = session.query(Album) \
                   .filter_by(title=album_name[:Album.TITLE_LIMIT]).one()
    assert album

    track_title = "*" * (Track.TITLE_LIMIT + 1)
    track = Track(lib_id=db_library.id, artist_id=artist.id, album_id=album.id)
    track.update(mp3audiofile)
    track.title = track_title
    session.add(track)
    session.commit()
    track = session.query(Track) \
                   .filter_by(title=track_title[:Track.TITLE_LIMIT]).one()
    assert track


def test_ImageMd5Validator(session, db_library, request, mp3audiofile):
    img = Image(type=Image.FRONT_COVER_TYPE, size=666,
                md5="48a6de198166ad9806f1a2172f2947df",
                mime_type="img/jpg", description="Iceburn",
                data=b"\xde\xad\xbe\xef")
    session.add(img)
    session.commit()
    img = session.query(Image).filter_by(data=b"\xde\xad\xbe\xef").one()
    assert img

    # Invalid md5
    with pytest.raises(ValueError):
        img.md5 = "Warzone"

    with pytest.raises(ValueError):
        img = Image(type=Image.FRONT_COVER_TYPE, size=666, md5="12345",
                    mime_type="img/sexy", description="Iceburn",
                    data=b"\xde\xad\xbe\xef")

def test_ImageLimits(session, db_library, request, mp3audiofile):
    desc = "d" * (Image.DESC_LIMIT + 1)
    img = Image(type=Image.FRONT_COVER_TYPE, size=666,
                md5="48a6de198166ad9806f1a2172f2947df",
                mime_type="img/png", description=desc,
                data=b"\xde\xad\xbe\xef")
    session.add(img)
    session.commit()
    img = session.query(Image).filter_by(description=desc[:Image.DESC_LIMIT])\
                 .one()
    assert img

def test_ImageType(session, db_library, request, mp3audiofile):
    img = Image(type="Beyond", size=666,
                md5="48a6de198166ad9806f1a2172f2947df",
                mime_type="img/jpg", description="Iceburn",
                data=b"\xde\xad\xbe\xef")
    session.add(img)

    if "sqlite" in request.keywords:
        with pytest.raises(IntegrityError):
            session.commit()
    else:
        assert "postgresql" in request.keywords
        with pytest.raises(DataError):
            session.commit()




import os
import platform
import time
import collections
from pathlib import Path
from os.path import getctime
from datetime import datetime

from nicfit import getLogger
from sqlalchemy.orm.exc import NoResultFound

import eyed3
import eyed3.main
from eyed3.utils import art
from eyed3.plugins import LoaderPlugin
from eyed3.utils.prompt import PromptExit
from eyed3.main import main as eyed3_main
from eyed3.core import TXXX_ALBUM_TYPE, VARIOUS_TYPE, LP_TYPE, SINGLE_TYPE, EP_TYPE
from nicfit.console.ansi import Fg
from nicfit.console import pout, perr

from ...util import normalizeCountry
from ...orm import (Track, Artist, Album, Meta, Image, Library,
                    VARIOUS_ARTISTS_ID, VARIOUS_ARTISTS_NAME, MAIN_LIB_NAME, NULL_LIB_ID)
from ... import console
from ... import database as db
from ...core import Command, EP_MAX_SIZE_HINT
from ...config import MusicLibrary

from .utils import syncImage, deleteOrphans

log = getLogger(__name__)
IMAGE_TYPES = {"artist": (Image.LOGO_TYPE, Image.ARTIST_TYPE, Image.LIVE_TYPE),
               "album": (Image.FRONT_COVER_TYPE, Image.BACK_COVER_TYPE,
                         Image.MISC_COVER_TYPE),
              }


class SyncPlugin(LoaderPlugin):
    """An eyeD3 file scanner/loader plugin."""

    NAMES = ['mishmash-sync']
    SUMMARY = "Synchronize files/directories with a Mishmash database."
    DESCRIPTION = ""

    def __init__(self, arg_parser):
        """Constructor"""
        super().__init__(arg_parser, cache_files=True, track_images=True)

        eyed3.main.setFileScannerOpts(
            arg_parser, default_recursive=True, paths_metavar="PATH_OR_LIB",
            paths_help="Files/directory paths, or individual music libraries. "
                       "No arguments will sync all configured libraries.")

        arg_parser.add_argument(
                "--monitor", action="store_true", dest="monitor",
                help="Monitor sync'd dirs for changes.")
        arg_parser.add_argument(
                "-f", "--force", action="store_true", dest="force",
                help="Force sync a library when sync=False.")
        arg_parser.add_argument(
                "--no-purge", action="store_true", dest="no_purge",
                help="Do not purge orphaned data (tracks, artists, albums, "
                     "etc.). This will make for a faster sync, and useful when "
                     "files were only added to a library.")
        arg_parser.add_argument(
                "--no-prompt", action="store_true", dest="no_prompt",
                help="Skip files that require user input.")
        arg_parser.add_argument(
            "--speed", default="fast", choices=("fast", "normal"),
            help="Sync speed. 'fast' will skips files whose timestamps have not changed, while "
                 "'normal' scans all files all the time.")

        self.monitor_proc = None
        self._num_added = 0
        self._num_modified = 0
        self._num_deleted = 0
        self._db_session = None
        self._lib = None
        self.start_time = None

    def start(self, args, config):
        import eyed3.utils.prompt

        self._num_loaded = 0

        eyed3.utils.prompt.DISABLE_PROMPT = "raise" if args.no_prompt else None

        super().start(args, config)
        self.start_time = time.time()
        self._db_session = args.db_session

        try:
            lib = self._db_session.query(Library)\
                                  .filter_by(name=args._library.name).one()
        except NoResultFound:
            lib = Library(name=args._library.name)
            self._db_session.add(lib)
            self._db_session.flush()
        self._lib = lib

        if self.args.monitor:
            from ._inotify import Monitor
            if self.monitor_proc is None:
                self.monitor_proc = Monitor()
            # Monitor roots, file dir are watched as the files are traversed
            for p in self.args.paths:
                self._watchDir(p)

    def _getArtist(self, session, name, origin, resolved_artist):
        origin_dict = {"origin_city": origin.city if origin else None,
                       "origin_state": origin.state if origin else None,
                       "origin_country": normalizeCountry(origin.country) if origin else None,
                      }
        if name == VARIOUS_ARTISTS_NAME:
            artist_rows = [session.query(Artist).filter_by(name=VARIOUS_ARTISTS_NAME,
                                                           lib_id=NULL_LIB_ID).one()]
        else:
            artist_rows = session.query(Artist)\
                                 .filter_by(name=name,
                                            lib_id=self._lib.id,
                                            **origin_dict)\
                                 .all()

        if artist_rows:
            if len(artist_rows) > 1 and resolved_artist:
                # Use previously resolved artist for this directory.
                artist = resolved_artist
            elif len(artist_rows) > 1:
                # Resolve artist
                try:
                    heading = "Multiple artists names '%s'" % \
                              artist_rows[0].name
                    artist = console.selectArtist(Fg.blue(heading),
                                                  choices=artist_rows,
                                                  allow_create=True)
                except PromptExit:
                    log.warning("Duplicate artist requires user intervention to resolve.")
                    artist = None
                else:
                    if artist not in artist_rows:
                        session.add(artist)
                        session.flush()
                        pout(Fg.blue("Updating artist") + ": " + name)
                    resolved_artist = artist
            else:
                assert len(artist_rows) == 1
                # Artist match
                artist = artist_rows[0]
        else:
            # New artist
            artist = Artist(name=name, lib_id=self._lib.id, **origin_dict)
            session.add(artist)
            session.flush()
            pout(Fg.green("Adding artist") + ": " + name)

        return artist, resolved_artist

    def _syncAudioFile(self, audio_file, album_type, d_datetime, session):
        path = audio_file.path
        info = audio_file.info
        tag = audio_file.tag

        album = None
        is_various = (album_type == VARIOUS_TYPE)

        if not info or not tag:
            log.warning(f"File missing {'audio' if not info else 'tag/metadata'}, skipping: {path}")
            return None, None
        elif None in (tag.title, tag.artist):
            log.warning("File missing required artist and/or title "
                        "metadata, skipping: %s" % path)
            return None, None

        # Used when a duplicate artist is resolved for the entire directory.
        resolved_artist = None
        resolved_album_artist = None

        try:
            track = session.query(Track)\
                           .filter_by(path=path, lib_id=self._lib.id).one()
        except NoResultFound:
            track = None
        else:
            if (self.args.speed == "fast"
                    and datetime.fromtimestamp(getctime(path)) == track.ctime):
                # Track is in DB and the file is not modified.
                return track, track.album

        # Either adding the track (track == None)
        # or modifying (track != None)

        artist, resolved_artist = self._getArtist(session, tag.artist, tag.artist_origin,
                                                  resolved_artist)
        if album_type != SINGLE_TYPE:
            if tag.album_artist and tag.artist != tag.album_artist:
                album_artist, resolved_album_artist = self._getArtist(session, tag.album_artist,
                                                                      tag.artist_origin,
                                                                      resolved_album_artist)
            else:
                album_artist = artist

            if artist is None:
                # see PromptExit
                return None, None

            album_artist_id = album_artist.id if not is_various \
                                              else VARIOUS_ARTISTS_ID
            rel_date = tag.release_date
            rec_date = tag.recording_date
            or_date = tag.original_release_date

            # Original release date
            if or_date:
                album = session.query(Album).filter_by(lib_id=self._lib.id,
                                                       artist_id=album_artist_id,
                                                       title=tag.album,
                                                       original_release_date=or_date).one_or_none()
            # Release date
            if not album and rel_date:
                album = session.query(Album).filter_by(lib_id=self._lib.id,
                                                       artist_id=album_artist_id,
                                                       title=tag.album,
                                                       release_date=rel_date).one_or_none()
            # Recording date
            if not album and rec_date:
                album = session.query(Album).filter_by(lib_id=self._lib.id,
                                                       artist_id=album_artist_id,
                                                       title=tag.album,
                                                       release_date=rel_date,
                                                       recording_date=rec_date).one_or_none()
            if album is None:
                album = Album(title=tag.album, lib_id=self._lib.id,
                              artist_id=album_artist_id, type=album_type,
                              release_date=rel_date,
                              original_release_date=or_date,
                              recording_date=rec_date,
                              date_added=d_datetime)
                pout(f"{Fg.green('Adding album')}: {album.title}")
                session.add(album)
            else:
                if album.type != album_type:
                    pout(Fg.yellow("Updating album") + ": " + album.title)
                    album.type = album_type

            session.flush()

        if not track:
            track = Track(audio_file=audio_file, lib_id=self._lib.id)
            self._num_added += 1
            pout(Fg.green("Adding track") + ": " + path)
        else:
            track.update(audio_file)
            self._num_modified += 1
            pout(Fg.yellow("Updating track") + ": " + path)

        track.artist_id = artist.id
        track.album_id = album.id if album else None

        if tag.genre:
            # Not uncommon for multiple genres to be 0x00 delimited
            for genre in tag.genre.name.split("\x00"):
                genre_tag = db.getTag(genre, session, self._lib.id, add=True)
                track.tags.append(genre_tag)

        session.add(track)

        if album:
            # Tag images
            img_type = None
            for img in tag.images:
                for img_type in art.TO_ID3_ART_TYPES:
                    if img.picture_type in art.TO_ID3_ART_TYPES[img_type]:
                        break
                    img_type = None

                if img_type is None:
                    log.warning(f"Skipping unsupported image type: {img.picture_type}")
                    continue

                new_img = Image.fromTagFrame(img, img_type)
                if new_img:
                    syncImage(new_img,
                              album if img_type in IMAGE_TYPES["album"]
                                    else album.artist,
                              session)
                else:
                    log.warning("Invalid image in tag")

        return track, album

    @staticmethod
    def _albumTypeHint(audio_files):
        types = collections.Counter()

        # This directory of files can be:
        # 1) an album by a single artist (tag.artist, or tag.album_artist and
        #    tag.album all equal)
        # 2) a comp (tag.album equal, tag.artist differ)
        # 3) not associated with a collection (tag.artist and tag.album differ)

        for tag in [f.tag for f in audio_files if f.tag]:
            album_type = tag.user_text_frames.get(TXXX_ALBUM_TYPE)
            if album_type:
                types[album_type.text] += 1

        if len(types) == 1:
            return types.most_common()[0][0]

        if len(types) == 0:
            artist_set = set()
            album_artist_set = set()
            albums = list()
            for tag in [f.tag for f in audio_files if f.tag]:
                if tag.artist:
                    artist_set.add(tag.artist)

                if tag.album_artist:
                    album_artist_set.add(tag.album_artist)

                if tag.album:
                    albums.append(tag.album)

            is_various = (
                len(artist_set) > 1
                and (len(album_artist_set) == 0 or album_artist_set == {VARIOUS_ARTISTS_NAME})
                and len(albums) == len(audio_files)
            )

            if is_various:
                return VARIOUS_TYPE
            elif len(albums) == len(audio_files):
                return LP_TYPE if len(audio_files) > EP_MAX_SIZE_HINT else EP_TYPE
            else:
                return SINGLE_TYPE

        if len(types) > 1:
            log.warning("Inconsistent type hints: %s" % str(types.keys()))
            return None

    def handleDirectory(self, d, _):
        pout(Fg.blue("Syncing directory") + ": " + str(d))
        audio_files = list(self._file_cache)
        self._file_cache = []

        image_files = self._dir_images
        self._dir_images = []

        if not audio_files:
            return

        d_datetime = datetime.fromtimestamp(getctime(d))

        album_type = self._albumTypeHint(audio_files) or LP_TYPE

        album = None
        session = self._db_session
        for audio_file in audio_files:
            try:
                track, album = self._syncAudioFile(audio_file, album_type, d_datetime, session)
            except Exception:
                # TODO: log and skip????
                raise

        if album:
            # Directory images.
            for img_file in image_files:
                img_type = art.matchArtFile(img_file)
                if img_type is None:
                    log.warning(f"Skipping unrecognized image file: {img_file}")
                    continue

                new_img = Image.fromFile(img_file, img_type)
                if new_img:
                    new_img.description = os.path.basename(img_file)
                    syncImage(new_img, album if img_type in IMAGE_TYPES["album"]
                                             else album.artist,
                              session)
                else:
                    log.warning(f"Invalid image file: {img_file}")

        session.commit()
        if self.args.monitor:
            self._watchDir(d)

    def _watchDir(self, d):
        valid_path = False
        dirpath = Path(d)
        for root in [Path(p) for p in self.args.paths]:
            try:
                dirpath.relative_to(root)
            except ValueError:
                continue
            else:
                valid_path = True
                self.monitor_proc.dir_queue.put((self._lib.name, dirpath))
                # Add parents up to root. It is safe to all the same dir to
                # the  Monitor
                parent = dirpath.parent
                try:
                    parent.relative_to(root)
                except ValueError:
                    # Added a root
                    pass
                else:
                    while parent != root:
                        self.monitor_proc.dir_queue.put((self._lib.name,
                                                         parent))
                        parent = parent.parent
                break
        assert valid_path

    def handleDone(self):
        t = time.time() - self.start_time
        session = self._db_session

        session.query(Meta).one().last_sync = datetime.utcnow()
        self._lib.last_sync = datetime.utcnow()

        num_orphaned_artists = 0
        num_orphaned_albums = 0
        if not self.args.no_purge:
            log.debug("Purging orphans (tracks, artists, albums) from database")
            (self._num_deleted,
             num_orphaned_artists,
             num_orphaned_albums) = deleteOrphans(session)

        if self._num_loaded or self._num_deleted:
            pout("")
            pout("== Library '{}' sync'd [ {:.2f}s time ({:.1f} files/sec) ] =="
                .format(self._lib.name, t, self._num_loaded / t))
            pout("%d files sync'd" % self._num_loaded)
            pout("%d tracks added" % self._num_added)
            pout("%d tracks modified" % self._num_modified)
            if not self.args.no_purge:
                pout("%d orphaned tracks deleted" % self._num_deleted)
                pout("%d orphaned artists deleted" % num_orphaned_artists)
                pout("%d orphaned albums deleted" % num_orphaned_albums)
            pout("")


@Command.register
class Sync(Command):
    NAME = "sync"
    HELP = "Synchronize music directories with database."

    def __init__(self, subparsers):
        super(Sync, self).__init__(subparsers)
        self.plugin = SyncPlugin(self.parser)
        self.args = None

    def _run(self, args=None):
        args = args or self.args
        args.plugin = self.plugin

        if args.monitor and platform.system() == "Darwin":
            perr("Monitor mode is not supported on OS/X\n")
            self.parser.print_usage()
            return 1

        libs = {lib.name: lib for lib in args.config.music_libs}
        if not libs and not args.paths:
            perr("\nMissing at least one path/library in which to sync!\n")
            self.parser.print_usage()
            return 1

        sync_libs = []
        if args.paths:
            file_paths = []
            for arg in args.paths:
                if arg in libs:
                    # Library name
                    sync_libs.append(libs[arg])
                else:
                    # Path
                    file_paths.append(arg)
            if file_paths:
                sync_libs.append(MusicLibrary(MAIN_LIB_NAME, paths=file_paths))
        else:
            sync_libs = list(libs.values())

        args.db_engine, args.db_session = self.db_engine, self.db_session

        def _syncLib(lib):
            args._library = lib

            args.paths = []
            for p in lib.paths:
                args.paths.append(str(p) if isinstance(p, Path) else p)

            pout("{}yncing library '{}': paths={}".format("Force s" if args.force else "S",
                                                          lib.name, lib.paths),
                 log=log)

            args.excludes = lib.excludes
            return eyed3_main(args, None)

        try:
            for lib in sync_libs:
                if not lib.sync and not args.force:
                    pout("[{}] - sync=False".format(lib.name), log=log)
                    continue
                result = _syncLib(lib)
                if result != 0:
                    return result
        except IOError as err:
            perr(str(err))
            return 1

        if args.monitor:
            from ._inotify import SYNC_INTERVAL
            monitor = self.plugin.monitor_proc

            # Commit now, since we won't be returning
            self.db_session.commit()

            monitor.start()
            try:
                while True:
                    if monitor.sync_queue.empty():
                        time.sleep(SYNC_INTERVAL / 2)
                        continue

                    sync_libs = {}
                    for i in range(monitor.sync_queue.qsize()):
                        lib, path = monitor.sync_queue.get_nowait()
                        if lib not in sync_libs:
                            sync_libs[lib] = set()
                        sync_libs[lib].add(path)

                    for lib, paths in sync_libs.items():
                        result = _syncLib(MusicLibrary(lib, paths=paths))
                        if result != 0:
                            return result
                        self.db_session.commit()
            finally:
                monitor.join()



import os
import argparse
from urllib.parse import urlparse
from eyed3.utils import datePicker
from countrycode.countrycode import countrycode

NAME_PREFIXES = ["the ", "los ", "la ", "el "]


def splitNameByPrefix(s):
    s_lower = s.lower()
    for prefix in NAME_PREFIXES:
        if s_lower.startswith(prefix):
            return (s[len(prefix):], s[0:len(prefix.rstrip())])
    return (s, None)


def sortByDate(things, prefer_recording_date=False):
    # XXX: Why just just make Album types sortable by intregating this
    def _sortkey(a):
        return datePicker(a, prefer_recording_date=prefer_recording_date) or 0
    return sorted(things, key=_sortkey)


def normalizeCountry(country_str, target="iso3c", title_case=False):
    """Return a normalized name/code for country in ``country_str``.
    The input can be a code or name, the ``target`` determines output value.
    3 character ISO code is the default (iso3c), 'country_name', and 'iso2c'
    are common also. See ``countrycode.countrycode`` for details and other
    options. Raises ``ValueError`` if the country is unrecognized."""
    iso2 = "iso2c"
    iso3 = "iso3c"
    raw = "country_name"

    if country_str is None:
        return ""

    if len(country_str) == 2:
        cc = countrycode(country_str.upper(), origin=iso2, target=target)
        if not cc:
            cc = countrycode(country_str, origin=raw, target=target)
    elif len(country_str) == 3:
        cc = countrycode(country_str.upper(), origin=iso3, target=target)
        if not cc:
            cc = countrycode(country_str, origin=raw, target=target)
    else:
        cc = countrycode(country_str, origin=raw, target=target)

    # Still need to validate because origin=raw will return whatever is
    # input if not match is found.
    cc = countrycode(cc, origin=target, target=target) if cc else None
    if not cc:
        raise ValueError("Country not found: %s" % (country_str))

    return cc.title() if title_case else cc


def commonDirectoryPrefix(*args):
    return os.path.commonprefix(args).rpartition(os.path.sep)[0]


def mostCommonItem(lst):
    """Choose the most common item from the list, or the first item if all
    items are unique."""
    # FIXME: Replace with collections.Counter
    # This elegant solution from: http://stackoverflow.com/a/1518632/1760218
    lst = [l for l in lst if l]
    if lst:
        return max(set(lst), key=lst.count)
    else:
        return None


def safeDbUrl(db_url):
    """Obfuscates password from a database URL."""
    url = urlparse(db_url)
    return db_url.replace(url.password, "****") if url.password else db_url


def addLibraryArguments(cli: argparse.ArgumentParser, nargs):
    """Add library options (-L/--library) with specific `nargs`."""
    from .orm import MAIN_LIB_NAME

    if nargs is not None:
        required = nargs in ("+", 1)
        if nargs not in ("?", 1):
            action, default, dest = ("append",
                                     [] if not required else [MAIN_LIB_NAME],
                                     "libs")
        else:
            action, default, dest = ("store",
                                     None if required else MAIN_LIB_NAME,
                                     "lib")

        cli.add_argument("-L", "--library", dest=dest, required=required,
                         action=action, metavar="LIB_NAME", default=default,
                         help="Specify a library.")



import os

from eyed3.core import VARIOUS_TYPE
from eyed3.utils.prompt import prompt
from nicfit.console.ansi import Style, Fg

from ..orm import Artist, Library, Image, IMAGE_TABLES
from ..core import Command
from ..console import promptArtist, selectArtist
from ..util import normalizeCountry, commonDirectoryPrefix, mostCommonItem

"""Metadata management commands."""


@Command.register
class SplitArtists(Command):
    NAME = "split-artists"
    HELP = "Split a single artist name into N distinct artists."
    _library_arg_nargs = 1

    def _initArgParser(self, parser):
        super()._initArgParser(parser)
        parser.add_argument("artist", help="The name of the artist to split.")

    def _displayArtistMusic(self, artist, albums, singles):
        if albums:
            print("%d albums by %s:" % (len(albums),
                                         Style.bright(Fg.blue(artist.name))))
            for alb in albums:
                print("%s %s" % (str(alb.getBestDate()).center(17),
                                  alb.title))

        if singles:
            print("%d single tracks by %s" %
                  (len(singles), Style.bright(Fg.blue(artist.name))))
            for s in singles:
                print("\t%s" % (s.title))

    def _run(self):
        session = self.db_session

        lib = session.query(Library).filter(Library.name == self.args.lib).one()
        artists = session.query(Artist).filter(Artist.lib_id == lib.id)\
                                       .filter(Artist.name == self.args.artist)\
                                       .all()
        if not artists:
            print("Artist not found: %s" % self.args.artist)
            return 1
        elif len(artists) > 1:
            artist = selectArtist(Fg.blue("Select which '%s' to split...") %
                                  artists[0].name,
                                  choices=artists, allow_create=False)
        else:
            artist = artists[0]

        # Albums by artist
        albums = list(artist.albums) + artist.getAlbumsByType(VARIOUS_TYPE)
        # Singles by artist and compilations the artist appears on
        singles = artist.getTrackSingles()

        if len(albums) < 2 and len(singles) < 2:
            print("%d albums and %d singles found for '%s', nothing to do." %
                    (len(albums), len(singles), artist.name))
            return 0

        self._displayArtistMusic(artist, albums, singles)

        def _validN(_n):
            try:
                return _n > 1 and _n <= len(albums)
            except Exception:
                return False

        n = prompt("\nEnter the number of distinct artists", type_=int,
                   validate=_validN)
        new_artists = []
        for i in range(1, n + 1):
            print(Style.bright("\n%s #%d") % (Fg.blue(artist.name), i))

            # Reuse original artist for first
            a = artist if i == 1 else Artist(name=artist.name,
                                             date_added=artist.date_added,
                                             lib_id=artist.lib_id)
            a.origin_city = prompt("   City", required=False)
            a.origin_state = prompt("   State", required=False)
            a.origin_country = prompt("   Country", required=False,
                                      type_=normalizeCountry)

            new_artists.append(a)

        if not Artist.checkUnique(new_artists):
            print(Fg.red("Artists must be unique."))
            return 1

        for a in new_artists:
            session.add(a)

        # New Artist objects need IDs
        session.flush()

        print(Style.bright("\nAssign albums to the correct artist."))
        for i, a in enumerate(new_artists):
            print("Enter %s%d%s for %s from %s%s%s" %
                  (Style.BRIGHT, i + 1, Style.RESET_BRIGHT,
                  a.name,
                  Style.BRIGHT, a.origin(country_code="iso3c",
                                         title_case=False),
                  Style.RESET_BRIGHT))

        # prompt for correct artists
        def _promptForArtist(_text):
            a = prompt(_text, type_=int,
                       choices=range(1, len(new_artists) + 1))
            return new_artists[a - 1]

        print("")
        for alb in albums:
            # Get some of the path to help the decision
            path = commonDirectoryPrefix(*[t.path for t in alb.tracks])
            path = os.path.join(*path.split(os.sep)[-2:])

            a = _promptForArtist("%s (%s)" % (alb.title, path))
            if alb.type != VARIOUS_TYPE:
                alb.artist_id = a.id
            for track in alb.tracks:
                if track.artist_id == artist.id:
                    track.artist_id = a.id

        print("")
        for track in singles:
            a = _promptForArtist(track.title)
            track.artist_id = a.id

        session.flush()


@Command.register
class MergeArtists(Command):
    NAME = "merge-artists"
    HELP = "Merge two or more artists into a single artist."
    _library_arg_nargs = 1

    def _initArgParser(self, parser):
        super()._initArgParser(parser)
        parser.add_argument("artists", nargs="+",
                            help="The artist names to merge.")

    def _run(self):
        session = self.db_session

        lib = session.query(Library).filter(Library.name == self.args.lib).one()

        merge_list = []
        for artist_arg in self.args.artists:
            artists = session.query(Artist)\
                             .filter(Artist.name == artist_arg)\
                             .filter(Artist.lib_id == lib.id).all()
            if len(artists) == 1:
                merge_list.append(artists[0])
            elif len(artists) > 1:
                merge_list += selectArtist(
                        Fg.blue("Select the artists to merge..."),
                        multiselect=True, choices=artists)

        if len(merge_list) > 1:
            # Reuse lowest id
            artist_ids = {a.id: a for a in merge_list}
            min_id = min(*artist_ids.keys())
            artist = artist_ids[min_id]

            mc = mostCommonItem
            new_artist = promptArtist(
                    "Merging %d artists into new artist..." % len(merge_list),
                    default_name=mc([a.name for a in merge_list]),
                    default_city=mc([a.origin_city for a in merge_list]),
                    default_state=mc([a.origin_state for a in merge_list]),
                    default_country=mc([a.origin_country for a in merge_list]),
                    artist=artist)
            new_artist.lib_id = lib.id
        else:
            print("Nothing to do, %s" %
                    ("artist not found" if not len(merge_list)
                                        else "only one artist found"))
            return 1

        assert(new_artist in merge_list)

        for artist in merge_list:
            if artist is new_artist:
                continue

            with session.no_autoflush:
                for alb in list(artist.albums):
                    if alb.type != VARIOUS_TYPE:
                        alb.artist_id = new_artist.id
                        artist.albums.remove(alb)
                        with session.no_autoflush:
                            new_artist.albums.append(alb)

                    for track in alb.tracks:
                        if track.artist_id == artist.id:
                            # gotta check in case alb is type various
                            track.artist_id = new_artist.id

                for track in artist.getTrackSingles():
                    track.artist_id = new_artist.id

            # flush to get new artist ids in sync before delete, otherwise
            # cascade happens.
            session.flush()
            session.delete(artist)

            session.flush()

        # FIXME: prompt for whether the tags should be updated with the new
        # FIXME: name if it is new.


@Command.register
class Images(Command):
    NAME = "image"
    HELP = "Image mgmt."
    _library_arg_nargs = 1

    def _initArgParser(self, parser):
        super()._initArgParser(parser)
        parser.add_argument("ids", nargs="+", help="The image IDs operate on.")
        parser.add_argument("--remove", action="store_true",
                            help="Remove images from the database.")

    def _run(self):
        for id_ in self.args.ids:
            image = self.db_session.query(Image).filter(Image.id == int(id_)).first()
            if not image:
                print(f"Image not found: {id_}")
                continue

            if self.args.remove:
                # For orm types for these, but clean jump tables
                for table in IMAGE_TABLES:
                    self.db_session.execute(f"DELETE FROM {table.name} where img_id={id_}")
                self.db_session.delete(image)
            else:
                print(image)

        self.db_session.commit()



"""
Object to relational database mappings for all tables.
"""
import os
import re
from hashlib import md5
from datetime import datetime
from gettext import gettext as _

import sqlalchemy as sql
from sqlalchemy.engine import Engine
from sqlalchemy.types import TypeDecorator
from sqlalchemy import orm, event, types, Sequence
from sqlalchemy.ext.declarative import declarative_base

import eyed3
from eyed3.utils import art
from eyed3.utils import guessMimetype
from eyed3.core import Date as Eyed3Date
from eyed3.core import ALBUM_TYPE_IDS, VARIOUS_TYPE, LIVE_TYPE
from eyed3.id3 import ID3_V1_0, ID3_V1_1, ID3_V2_2, ID3_V2_3, ID3_V2_4, versionToString

VARIOUS_ARTISTS_ID = 1
VARIOUS_ARTISTS_NAME = _("Various Artists")
NULL_LIB_ID = 1
NULL_LIB_NAME = "__null_lib__"
MAIN_LIB_ID = 2
MAIN_LIB_NAME = _("Music")
DATE_LIMIT = 24

convention = {
  "ix": 'ix_%(column_0_label)s',
  "uq": "uq_%(table_name)s_%(column_0_name)s",
  "ck": "ck_%(table_name)s_%(constraint_name)s",
  "fk": "fk_%(table_name)s_%(column_0_name)s_%(referred_table_name)s",
  "pk": "pk_%(table_name)s"
}

Base = declarative_base(metadata=sql.MetaData(naming_convention=convention))

artist_tags = sql.Table("artist_tags", Base.metadata,
                        sql.Column("artist_id", sql.Integer,
                                   sql.ForeignKey("artists.id")),
                        sql.Column("tag_id", sql.Integer,
                                   sql.ForeignKey("tags.id")),
                         )
"""Pivot table 'artist_tags' for mapping an artist ID to a value in the
`tags` table."""

album_tags = sql.Table("album_tags", Base.metadata,
                       sql.Column("album_id", sql.Integer,
                                  sql.ForeignKey("albums.id")),
                       sql.Column("tag_id", sql.Integer,
                                  sql.ForeignKey("tags.id")),
                        )
"""Pivot table 'album_tags' for mapping an album ID to a value in the
`tags` table."""

track_tags = sql.Table("track_tags", Base.metadata,
                       sql.Column("track_id", sql.Integer,
                                  sql.ForeignKey("tracks.id")),
                       sql.Column("tag_id", sql.Integer,
                                  sql.ForeignKey("tags.id")),
                        )
"""Pivot table 'track_tags' for mapping a track ID to a value in the
`tags` table."""

artist_images = sql.Table("artist_images", Base.metadata,
                          sql.Column("artist_id", sql.Integer,
                                     sql.ForeignKey("artists.id")),
                          sql.Column("img_id", sql.Integer,
                                     sql.ForeignKey("images.id")),
                         )
"""Pivot table 'artist_images' for mapping an artist ID to a value in the
`images` table."""

album_images = sql.Table("album_images", Base.metadata,
                         sql.Column("album_id", sql.Integer,
                                    sql.ForeignKey("albums.id")),
                         sql.Column("img_id", sql.Integer,
                                    sql.ForeignKey("images.id")),
                        )
"""Pivot table 'album_images' for mapping an album ID to a value in the
`images` table."""


class OrmObject(object):
    """Base classes for all other mishmash.orm classes."""

    @staticmethod
    def _truncate(string, limit):
        # TODO: Append invisible marker to signal truncation as to NOT reflect
        # TODO: in the original source
        if string and len(string) > limit:
            string = string[:limit]
        return string

    def __repr__(self):  # pragma: nocover
        """Dump the object state and return it as a strings."""
        attrs = []
        for key in self.__dict__:
            if not key.startswith('_'):
                attrs.append((key, getattr(self, key)))
        return self.__class__.__name__ + '(' + \
               ', '.join(x[0] + '=' + repr(x[1]) for x in attrs) + ')'


class Meta(Base, OrmObject):
    """Table ``meta`` used for storing database schema version, timestamps,
    and any other metadata about the music collection."""
    __tablename__ = "meta"
    VERSION_LIMIT = 32

    # Columns
    version = sql.Column(sql.String(VERSION_LIMIT), nullable=False,
                         primary_key=True)
    """The MishMash version defines the database schema."""
    last_sync = sql.Column(sql.DateTime)
    """A UTC timestamp of the last sync operation."""


def getSortName(name):
    from ..util import splitNameByPrefix
    suffix, prefix = splitNameByPrefix(name)
    return "%s, %s" % (suffix, prefix) if prefix else name


class Artist(Base, OrmObject):
    __tablename__ = "artists"
    __table_args__ = (sql.UniqueConstraint("name",
                                           "origin_city",
                                           "origin_state",
                                           "origin_country",
                                           "lib_id",
                                          ), {})

    NAME_LIMIT = 256
    SORT_NAME_LIMIT = NAME_LIMIT + 2

    CITY_LIMIT = 64
    STATE_LIMIT = 32
    COUNTRY_LIMIT = 3

    # Columns
    id = sql.Column(sql.Integer, Sequence("artists_id_seq"), primary_key=True)
    name = sql.Column(sql.Unicode(NAME_LIMIT), nullable=False, index=True)
    sort_name = sql.Column(sql.Unicode(SORT_NAME_LIMIT), nullable=False)
    date_added = sql.Column(sql.DateTime(), nullable=False,
                            default=datetime.now)
    origin_city = sql.Column(sql.Unicode(CITY_LIMIT))
    origin_state = sql.Column(sql.Unicode(STATE_LIMIT))
    origin_country = sql.Column(sql.String(COUNTRY_LIMIT))

    # Foreign keys
    lib_id = sql.Column(sql.Integer, sql.ForeignKey("libraries.id"),
                        nullable=False, index=True)

    # Relations
    albums = orm.relation("Album", cascade="all")
    """all albums by the artist"""
    tracks = orm.relation("Track", cascade="all")
    """all tracks by the artist"""
    tags = orm.relation("Tag", secondary=artist_tags)
    """one-to-many (artist->tag) and many-to-one (tag->artist)"""
    images = orm.relation("Image", secondary=artist_images, cascade="all")
    """one-to-many artist images."""
    library = orm.relation("Library")

    def getAlbumsByType(self, album_type):
        if album_type == VARIOUS_TYPE:
            albums = set([t.album for t in self.tracks
                                  if t.album and t.album.type == album_type])
            albums = list(albums)
        else:
            albums = [a for a in self.albums if a.type == album_type]

        return albums

    def getTrackSingles(self):
        tracks = []
        for t in self.tracks:
            if t.album_id is None:
                # Include single files, not associated with an album
                tracks.append(t)
            elif t.album and t.album.artist_id != self.id:
                # Include tracks that the artist appears on.
                tracks.append(t)

        return tracks

    @property
    def url_name(self):
        return self.name.replace("/", "%2f")

    def origin(self, n=3, country_code="country_name", title_case=True):
        from ..util import normalizeCountry
        origins = [o for o in [normalizeCountry(self.origin_country,
                                                target=country_code,
                                                title_case=title_case),
                               self.origin_state,
                               self.origin_city]
                     if o]
        origins = origins[:n]
        origins.reverse()
        return ", ".join(origins)

    @orm.validates("name")
    def _setName(self, key, value):
        if not value:
            raise ValueError("Artist.name is not nullable")
        value = self._truncate(value, self.NAME_LIMIT)
        self.sort_name = getSortName(value)
        return value

    @orm.validates("origin_country")
    def _setOriginCountry(self, _, value):
        from ..util import normalizeCountry
        if value is None:
            return None
        return normalizeCountry(value, target="iso3c", title_case=False)

    @orm.validates("origin_city", "origin_state")
    def _truncateOrigins(self, key, value):
        limit = self.CITY_LIMIT if key == "origin_city" else self.STATE_LIMIT
        return self._truncate(value, limit)

    @staticmethod
    def checkUnique(artists):
        vals = []
        for a in artists:
            v = (a.name, a.origin_city, a.origin_state, a.origin_country)
            if v in vals:
                return False
            vals.append(v)
        return True

    @property
    def is_various_artist(self):
        return self.id == VARIOUS_ARTISTS_ID


class AlbumDate(TypeDecorator):
    """Custom column type for eyed3.core.Date objects. That is, dates than
    can have empty rather than default date fields. For example, 1994 with no
    month and day is different than 1994-01-01, as datetime provides."""
    impl = types.String(DATE_LIMIT)

    def process_bind_param(self, value, dialect):
        if isinstance(value, Eyed3Date):
            return str(value)
        elif value:
            return str(Eyed3Date.parse(value))
        else:
            return None

    def process_result_value(self, value, dialect):
        return Eyed3Date.parse(value) if value else None


class Album(Base, OrmObject):
    __tablename__ = "albums"
    __table_args__ = (sql.UniqueConstraint("title",
                                           "artist_id",
                                           "lib_id",
                                           "release_date",
                                           "recording_date",
                                           "original_release_date",
                                          ), {})

    TITLE_LIMIT = 256
    DATE_LIMIT = DATE_LIMIT
    _types_enum = sql.Enum(*ALBUM_TYPE_IDS, name="album_types")

    # Columns
    id = sql.Column(sql.Integer, Sequence("albums_id_seq"), primary_key=True)
    title = sql.Column(sql.Unicode(TITLE_LIMIT), nullable=False, index=True)
    type = sql.Column(_types_enum, nullable=False, default=ALBUM_TYPE_IDS[0])
    date_added = sql.Column(sql.DateTime(), nullable=False,
                            default=datetime.now)
    release_date = sql.Column(AlbumDate)
    original_release_date = sql.Column(AlbumDate)
    recording_date = sql.Column(AlbumDate)

    # Foreign keys
    artist_id = sql.Column(sql.Integer, sql.ForeignKey("artists.id"),
                           nullable=False, index=True)
    lib_id = sql.Column(sql.Integer, sql.ForeignKey("libraries.id"),
                        nullable=False, index=True)

    # Relations
    artist = orm.relation("Artist")
    tracks = orm.relation("Track", order_by="Track.track_num",
                          cascade="all")
    tags = orm.relation("Tag", secondary=album_tags)
    images = orm.relation("Image", secondary=album_images, cascade="all")
    """one-to-many album images."""
    library = orm.relation("Library")

    def getBestDate(self):
        from eyed3.utils import datePicker
        return datePicker(self,
                          prefer_recording_date=bool(self.type == LIVE_TYPE))

    @property
    def duration(self):
        return sum([t.time_secs for t in self.tracks])

    @orm.validates("title")
    def _truncateTitle(self, key, value):
        return self._truncate(value, self.TITLE_LIMIT)


class Track(Base, OrmObject):
    __tablename__ = "tracks"
    __table_args__ = (sql.UniqueConstraint("path",
                                           "lib_id",
                                          ), {})

    METADATA_FORMATS = ["ID3{}".format(versionToString(v))
                            for v in (ID3_V1_0, ID3_V1_1, ID3_V2_2, ID3_V2_3,
                                      ID3_V2_4)]
    _metadata_enum = sql.Enum(*METADATA_FORMATS, name="metadata_format")
    PATH_LIMIT = 2048
    TITLE_LIMIT = 256

    # Columns
    id = sql.Column(sql.Integer, Sequence("tracks_id_seq"), primary_key=True)
    path = sql.Column(sql.String(PATH_LIMIT), nullable=False, index=True)
    size_bytes = sql.Column(sql.Integer, nullable=False)
    ctime = sql.Column(sql.DateTime(), nullable=False)
    mtime = sql.Column(sql.DateTime(), nullable=False)
    date_added = sql.Column(sql.DateTime(), nullable=False,
                            default=datetime.now)
    time_secs = sql.Column(sql.Float, nullable=False)
    title = sql.Column(sql.Unicode(TITLE_LIMIT), nullable=False, index=True)
    track_num = sql.Column(sql.SmallInteger)
    track_total = sql.Column(sql.SmallInteger)
    media_num = sql.Column(sql.SmallInteger)
    media_total = sql.Column(sql.SmallInteger)
    bit_rate = sql.Column(sql.SmallInteger)
    variable_bit_rate = sql.Column(sql.Boolean)
    metadata_format = sql.Column(_metadata_enum, nullable=False)

    # Foreign keys
    artist_id = sql.Column(sql.Integer, sql.ForeignKey("artists.id"),
                           nullable=False, index=True)
    album_id = sql.Column(sql.Integer, sql.ForeignKey("albums.id"),
                          nullable=True, index=True)
    lib_id = sql.Column(sql.Integer, sql.ForeignKey("libraries.id"),
                        nullable=False, index=True)

    # Relations
    artist = orm.relation("Artist")
    album = orm.relation("Album")
    tags = orm.relation("Tag", secondary=track_tags)
    library = orm.relation("Library")

    # XXX: test-only
    _mp3_file = None

    def __init__(self, **kwargs):
        """Along with the column args a ``audio_file`` keyword may be passed
        for this class to use for initialization."""

        if "audio_file" in kwargs:
            self.update(kwargs["audio_file"])
            del kwargs["audio_file"]

        super(Track, self).__init__(**kwargs)

    def update(self, audio_file):
        path = audio_file.path
        tag = audio_file.tag
        info = audio_file.info

        self.path = path
        self.size_bytes = info.size_bytes
        self.ctime = datetime.fromtimestamp(os.path.getctime(path))
        self.mtime = datetime.fromtimestamp(os.path.getmtime(path))
        self.time_secs = info.time_secs
        self.title = (tag.title if "\x00" not in tag.title
                                else tag.title.split("\x00")[0])
        self.track_num, self.track_total = tag.track_num
        self.variable_bit_rate, self.bit_rate = info.bit_rate
        self.media_num, self.media_total = tag.disc_num
        self.metadata_format = "ID3{}".format(versionToString(tag.version))

        return self

    @orm.validates("title")
    def _truncateTitle(self, _, value):
        return self._truncate(value, Track.TITLE_LIMIT)

    def loadAudioFile(self, path_mapping=None):
        path = self.path

        path_mapping = path_mapping or {}
        for fm, to in path_mapping.items():
            if path.startswith(fm):
                # Remap path
                path = to + path[len(fm):]

        return eyed3.load(path)


class Tag(Base, OrmObject):
    __tablename__ = "tags"
    __table_args__ = (sql.UniqueConstraint("name",
                                           "lib_id",
                                          ), {})

    NAME_LIMIT = 64

    # Columns
    id = sql.Column(sql.Integer, Sequence("tags_id_seq"), primary_key=True)
    name = sql.Column(sql.Unicode(NAME_LIMIT), nullable=False, unique=False)

    # Foreign keys
    lib_id = sql.Column(sql.Integer, sql.ForeignKey("libraries.id"),
                        nullable=False, index=True)

    # Relations
    library = orm.relation("Library")

    @orm.validates("name")
    def _truncateName(self, key, value):
        return self._truncate(value, self.NAME_LIMIT)


class Image(Base, OrmObject):
    __tablename__ = "images"

    FRONT_COVER_TYPE = art.FRONT_COVER
    BACK_COVER_TYPE = art.BACK_COVER
    MISC_COVER_TYPE = art.MISC_COVER
    LOGO_TYPE = art.LOGO
    ARTIST_TYPE = art.ARTIST
    LIVE_TYPE = art.LIVE
    IMAGE_TYPES = [FRONT_COVER_TYPE, BACK_COVER_TYPE, MISC_COVER_TYPE,
                   LOGO_TYPE, ARTIST_TYPE, LIVE_TYPE]
    MIMETYPE_LIMIT = 32
    MD5_LIMIT = 32
    DESC_LIMIT = 1024
    _types_enum = sql.Enum(*IMAGE_TYPES, name="image_types")

    id = sql.Column(sql.Integer, Sequence("images_id_seq"), primary_key=True)
    type = sql.Column(_types_enum, nullable=False)
    mime_type = sql.Column(sql.String(MIMETYPE_LIMIT), nullable=False)
    md5 = sql.Column(sql.String(MD5_LIMIT), nullable=False)
    size = sql.Column(sql.Integer, nullable=False)
    description = sql.Column(sql.String(DESC_LIMIT), nullable=False)
    """The description will be the base file name when the source if a file."""
    data = orm.deferred(sql.Column(sql.LargeBinary, nullable=False))

    @staticmethod
    def _validMimeType(mt):
        try:
            p1, p2 = mt.split("/")
            return p1 == "image" and p2
        except ValueError:
            # Missing '/'
            return False

    @staticmethod
    def fromTagFrame(img, type_):
        if not Image._validMimeType(img.mime_type):
            return None

        md5hash = md5()
        md5hash.update(img.image_data)

        return Image(type=type_,
                     description=img.description,
                     mime_type=img.mime_type,
                     md5=md5hash.hexdigest(),
                     size=len(img.image_data),
                     data=img.image_data)

    @staticmethod
    def fromFile(path, type_):
        mime_type = guessMimetype(path)
        if not Image._validMimeType(mime_type):
            return None
        md5hash = md5()

        img = open(path, "rb")
        data = img.read()
        img.close()

        md5hash.update(data)

        return Image(type=type_,
                     mime_type=guessMimetype(path),
                     md5=md5hash.hexdigest(),
                     size=len(data),
                     data=data)

    @orm.validates("md5")
    def _validMd5(self, key, value):
        if not re.compile(r"([a-fA-F\d]{32})$").match(value):
            raise ValueError(f"Invalid md5sum: {value}")
        return value

    @orm.validates("description")
    def _validLimits(self, key, value):
        return self._truncate(value, self.DESC_LIMIT)


class Library(Base, OrmObject):
    __tablename__ = "libraries"
    NAME_LIMIT = 64

    # Columns
    id = sql.Column(sql.Integer, Sequence("libraries_id_seq"), primary_key=True)
    name = sql.Column(sql.Unicode(NAME_LIMIT), nullable=False, unique=True)
    last_sync = sql.Column(sql.DateTime)

    _music = []

    def add(self, album):
        self._music.append(album)
        return self

    def albums(self):
        for alb in self._music:
            yield alb

    @classmethod
    def iterall(Class, session, names=None):
        """Iterate over all Library rows found in `session`.
        :param names: Optional sequence of names to filter on.
        """
        names = set(names if names else [])
        for lib in session.query(Class).filter(Class.id > NULL_LIB_ID).all():
            if not names or (lib.name in names):
                yield lib


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    """Allows foreign keys to work in sqlite."""
    import sqlite3
    if dbapi_connection.__class__ is sqlite3.Connection:
        cursor = dbapi_connection.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")
        cursor.close()



# XXX: revist, but this going away is nice
import warnings
warnings.filterwarnings("ignore", message="The psycopg2 wheel package will be renamed")

from nicfit import getLogger                                               # noqa: E402
from .orm.core import VARIOUS_ARTISTS_NAME                                 # noqa: E402
from .__about__ import __version__ as version                              # noqa: E402

log = getLogger(__package__)


__all__ = ["log", "getLogger", "version", "VARIOUS_ARTISTS_NAME"]



import sqlalchemy as sql
from . import core

USERNAME_LIMIT = 64
EMAIL_LIMIT = USERNAME_LIMIT + 128
PASSWORD_LIMIT = 128


class User(core.Base, core.OrmObject):
    __tablename__ = "users"

    id = sql.Column(sql.Integer, sql.Sequence("users_id_seq"), primary_key=True)
    username = sql.Column(sql.String(USERNAME_LIMIT), nullable=False, index=True)
    email = sql.Column(sql.String(EMAIL_LIMIT), nullable=False)
    password = sql.Column(sql.String(PASSWORD_LIMIT), nullable=False)



"""initial autogen revision

Revision ID: 74df8fa5a35f
Revises:
Create Date: 2017-02-06 20:48:47.191133

"""
import datetime
from alembic import op
import sqlalchemy as sa
import mishmash
from mishmash.orm import Library, Artist, Album, Track, Image, Tag, Meta, ALBUM_TYPE_IDS
from mishmash.orm import AlbumDate


# revision identifiers, used by Alembic.
revision = '74df8fa5a35f'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('images',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('type', sa.Enum('FRONT_COVER',
                                              'BACK_COVER',
                                              'MISC_COVER',
                                              'LOGO',
                                              'ARTIST',
                                              'LIVE',
                                              name='image_types'),
                              nullable=False),
                    sa.Column('mime_type',
                              sa.String(length=Image.MIMETYPE_LIMIT),
                              nullable=False),
                    sa.Column('md5', sa.String(length=Image.MD5_LIMIT),
                              nullable=False),
                    sa.Column('size', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=Image.DESC_LIMIT),
                              nullable=False),
                    sa.Column('data', sa.LargeBinary(), nullable=False),
                    sa.PrimaryKeyConstraint('id', name=op.f('pk_images'))
    )

    lib_t = op.create_table('libraries',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('name', sa.Unicode(length=Library.NAME_LIMIT),
                              nullable=False),
                    sa.Column('last_sync', sa.DateTime(), nullable=True),
                    sa.PrimaryKeyConstraint('id', name=op.f('pk_libraries')),
                    sa.UniqueConstraint('name', name=op.f('uq_libraries_name'))
    )
    op.bulk_insert(lib_t,
                   [
                       {"name": mishmash.orm.NULL_LIB_NAME},
                       {"name": mishmash.orm.MAIN_LIB_NAME},
                   ])

    meta_t = op.create_table('meta',
                    sa.Column('version',
                              sa.String(length=Meta.VERSION_LIMIT),
                              nullable=False),
                    sa.Column('last_sync', sa.DateTime(), nullable=True),
                    sa.PrimaryKeyConstraint('version', name=op.f('pk_meta'))
    )
    op.bulk_insert(meta_t, [{"version": mishmash.version}])

    artists_t = op.create_table('artists',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('name', sa.Unicode(length=Artist.NAME_LIMIT),
                              nullable=False),
                    sa.Column('sort_name',
                              sa.Unicode(length=Artist.SORT_NAME_LIMIT),
                              nullable=False),
                    sa.Column('date_added', sa.DateTime(), nullable=False),
                    sa.Column('origin_city',
                              sa.Unicode(length=Artist.CITY_LIMIT),
                              nullable=True),
                    sa.Column('origin_state',
                              sa.Unicode(length=Artist.STATE_LIMIT),
                              nullable=True),
                    sa.Column('origin_country',
                              sa.String(length=Artist.COUNTRY_LIMIT),
                              nullable=True),
                    sa.Column('lib_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(
                        ['lib_id'], ['libraries.id'],
                        name=op.f('fk_artists_lib_id_libraries')),
                    sa.PrimaryKeyConstraint('id', name=op.f('pk_artists')),
                    sa.UniqueConstraint('name', 'origin_city', 'origin_state',
                                        'origin_country', 'lib_id',
                                        name=op.f('uq_artists_name'))
    )
    op.bulk_insert(artists_t,
                   [{"lib_id": mishmash.orm.NULL_LIB_ID,
                     "name": mishmash.orm.VARIOUS_ARTISTS_NAME,
                     "sort_name": mishmash.orm.getSortName(
                                    mishmash.orm.VARIOUS_ARTISTS_NAME),
                     "date_added": datetime.datetime.now(),
                    }])

    op.create_index(op.f('ix_artists_lib_id'), 'artists', ['lib_id'],
                    unique=False)
    op.create_index(op.f('ix_artists_name'), 'artists', ['name'],
                    unique=False)
    op.create_table('tags',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('name', sa.Unicode(length=Tag.NAME_LIMIT),
                              nullable=False),
                    sa.Column('lib_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(
                        ['lib_id'], ['libraries.id'],
                        name=op.f('fk_tags_lib_id_libraries')),
                    sa.PrimaryKeyConstraint('id', name=op.f('pk_tags')),
                    sa.UniqueConstraint('name', 'lib_id',
                                        name=op.f('uq_tags_name'))
    )
    op.create_index(op.f('ix_tags_lib_id'), 'tags', ['lib_id'], unique=False)
    op.create_table('albums',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('title', sa.Unicode(length=Album.TITLE_LIMIT),
                              nullable=False),
                    sa.Column('type', sa.Enum(*ALBUM_TYPE_IDS, name='album_types'),
                              nullable=False),
                    sa.Column('date_added', sa.DateTime(), nullable=False),
                    sa.Column('release_date',
                              AlbumDate(length=Album.DATE_LIMIT),
                              nullable=True),
                    sa.Column('original_release_date',
                              AlbumDate(length=Album.DATE_LIMIT),
                              nullable=True),
                    sa.Column('recording_date',
                              AlbumDate(length=Album.DATE_LIMIT),
                              nullable=True),
                    sa.Column('artist_id', sa.Integer(), nullable=False),
                    sa.Column('lib_id', sa.Integer(), nullable=False),
                    sa.ForeignKeyConstraint(
                        ['artist_id'], ['artists.id'],
                        name=op.f('fk_albums_artist_id_artists')),
                    sa.ForeignKeyConstraint(
                        ['lib_id'], ['libraries.id'],
                        name=op.f('fk_albums_lib_id_libraries')),
                    sa.PrimaryKeyConstraint('id', name=op.f('pk_albums')),
                    sa.UniqueConstraint('title', 'artist_id', 'lib_id',
                                        'release_date', 'recording_date', 'original_release_date',
                                        name=op.f('uq_albums_title'))
    )
    op.create_index(op.f('ix_albums_artist_id'), 'albums', ['artist_id'],
                    unique=False)
    op.create_index(op.f('ix_albums_lib_id'), 'albums', ['lib_id'],
                    unique=False)
    op.create_index(op.f('ix_albums_title'), 'albums', ['title'],
                    unique=False)
    op.create_table('artist_images',
                    sa.Column('artist_id', sa.Integer(), nullable=True),
                    sa.Column('img_id', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(
                        ['artist_id'], ['artists.id'],
                        name=op.f('fk_artist_images_artist_id_artists')),
                    sa.ForeignKeyConstraint(
                        ['img_id'], ['images.id'],
                        name=op.f('fk_artist_images_img_id_images'))
    )
    op.create_table('artist_tags',
                    sa.Column('artist_id', sa.Integer(), nullable=True),
                    sa.Column('tag_id', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['artist_id'], ['artists.id'],
                            name=op.f('fk_artist_tags_artist_id_artists')),
    sa.ForeignKeyConstraint(['tag_id'], ['tags.id'],
                            name=op.f('fk_artist_tags_tag_id_tags'))
    )
    op.create_table('album_images',
                    sa.Column('album_id', sa.Integer(), nullable=True),
                    sa.Column('img_id', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(
                        ['album_id'], ['albums.id'],
                        name=op.f('fk_album_images_album_id_albums')),
                    sa.ForeignKeyConstraint(
                        ['img_id'], ['images.id'],
                        name=op.f('fk_album_images_img_id_images'))
    )
    op.create_table('album_tags',
                    sa.Column('album_id', sa.Integer(), nullable=True),
                    sa.Column('tag_id', sa.Integer(), nullable=True),
                    sa.ForeignKeyConstraint(
                        ['album_id'], ['albums.id'],
                        name=op.f('fk_album_tags_album_id_albums')),
                    sa.ForeignKeyConstraint(
                        ['tag_id'], ['tags.id'],
                        name=op.f('fk_album_tags_tag_id_tags'))
    )
    op.create_table('tracks',
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('path', sa.String(length=Track.PATH_LIMIT),
                              nullable=False),
                    sa.Column('size_bytes', sa.Integer(), nullable=False),
                    sa.Column('ctime', sa.DateTime(), nullable=False),
                    sa.Column('mtime', sa.DateTime(), nullable=False),
                    sa.Column('date_added', sa.DateTime(), nullable=False),
                    sa.Column('time_secs', sa.Float(), nullable=False),
                    sa.Column('title', sa.Unicode(length=Track.TITLE_LIMIT),
                              nullable=False),
                    sa.Column('track_num', sa.SmallInteger(), nullable=True),
                    sa.Column('track_total', sa.SmallInteger(), nullable=True),
                    sa.Column('media_num', sa.SmallInteger(), nullable=True),
                    sa.Column('media_total', sa.SmallInteger(), nullable=True),
                    sa.Column('bit_rate', sa.SmallInteger(), nullable=True),
                    sa.Column('variable_bit_rate', sa.Boolean(name="vbr_bool"),
                              nullable=True),
                    sa.Column('artist_id', sa.Integer(), nullable=False),
                    sa.Column('album_id', sa.Integer(), nullable=True),
                    sa.Column('lib_id', sa.Integer(), nullable=False),
                    sa.Column('metadata_format',
                              sa.Enum(*['ID3v1.0', 'ID3v1.1', 'ID3v2.2',
                                        'ID3v2.3', 'ID3v2.4'],
                                      name='metadata_formats')),
                    sa.ForeignKeyConstraint(
                        ['album_id'], ['albums.id'],
                        name=op.f('fk_tracks_album_id_albums')),
                    sa.ForeignKeyConstraint(
                        ['artist_id'], ['artists.id'],
                        name=op.f('fk_tracks_artist_id_artists')),
                    sa.ForeignKeyConstraint(
                        ['lib_id'], ['libraries.id'],
                        name=op.f('fk_tracks_lib_id_libraries')),
                    sa.PrimaryKeyConstraint('id', name=op.f('pk_tracks')),
                    sa.UniqueConstraint('path', 'lib_id',
                                        name=op.f('uq_tracks_path'))
    )
    op.create_index(op.f('ix_tracks_album_id'), 'tracks', ['album_id'],
                    unique=False)
    op.create_index(op.f('ix_tracks_artist_id'), 'tracks', ['artist_id'],
                    unique=False)
    op.create_index(op.f('ix_tracks_lib_id'), 'tracks', ['lib_id'],
                    unique=False)
    op.create_index(op.f('ix_tracks_path'), 'tracks', ['path'], unique=False)
    op.create_index(op.f('ix_tracks_title'), 'tracks', ['title'], unique=False)
    op.create_table('track_tags',
                    sa.Column('track_id', sa.Integer(), nullable=True),
                    sa.Column('tag_id', sa.Integer(), nullable=True),
                              sa.ForeignKeyConstraint(
                                  ['tag_id'], ['tags.id'],
                                  name=op.f('fk_track_tags_tag_id_tags')),
                              sa.ForeignKeyConstraint(
                                  ['track_id'], ['tracks.id'],
                                  name=op.f('fk_track_tags_track_id_tracks'))
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('track_tags')
    op.drop_index(op.f('ix_tracks_title'), table_name='tracks')
    op.drop_index(op.f('ix_tracks_path'), table_name='tracks')
    op.drop_index(op.f('ix_tracks_lib_id'), table_name='tracks')
    op.drop_index(op.f('ix_tracks_artist_id'), table_name='tracks')
    op.drop_index(op.f('ix_tracks_album_id'), table_name='tracks')
    op.drop_table('tracks')
    op.drop_table('album_tags')
    op.drop_table('album_images')
    op.drop_table('artist_tags')
    op.drop_table('artist_images')
    op.drop_index(op.f('ix_albums_title'), table_name='albums')
    op.drop_index(op.f('ix_albums_lib_id'), table_name='albums')
    op.drop_index(op.f('ix_albums_artist_id'), table_name='albums')
    op.drop_table('albums')
    op.drop_index(op.f('ix_tags_lib_id'), table_name='tags')
    op.drop_table('tags')
    op.drop_index(op.f('ix_artists_name'), table_name='artists')
    op.drop_index(op.f('ix_artists_lib_id'), table_name='artists')
    op.drop_table('artists')
    op.drop_table('meta')
    op.drop_table('libraries')
    op.drop_table('images')
    # ### end Alembic commands ###



""""""
from nicfit.console.ansi import Fg
from eyed3.utils.prompt import prompt, parseIntList
from .orm import Artist


def selectArtist(heading, choices=None, multiselect=False, allow_create=True):
    color = Fg.green
    artist = None
    name = None
    menu_num = 0

    if heading:
        print(heading)

    while artist is None:
        if choices:
            name = choices[0].name
            for menu_num, a in enumerate(choices, start=1):
                print("   %d) %s" % (menu_num + 1, a.origin()))

            if not multiselect:
                if allow_create:
                    menu_num += 1
                    print("   %d) Enter a new artist" % menu_num)

                choice = prompt("Which artist", type_=int,
                                choices=range(1, menu_num + 1))
                choice -= 1
                if choice < len(choices):
                    artist = choices[choice]
                # Otherwise fall through to select artist below
            else:
                def _validate(_resp):
                    try:
                        _ints = [_i for _i in parseIntList(_resp)
                                    if _i in range(1, menu_num + 1)]
                        return bool(_ints)
                    except Exception:
                        return False

                resp = prompt(color("Choose one or more artists"),
                              validate=_validate)
                artists = []
                for choice in [i - 1 for i in parseIntList(resp)]:
                    artists.append(choices[choice])
                # XXX: blech, returning a list here and a single value below
                return artists

        if artist is None:
            artist = promptArtist(None, name=name)
            if choices:
                if not Artist.checkUnique(choices + [artist]):
                    print(Fg.red("Artist entered is not unique, try again..."))
                    artist = None

    return artist


def promptArtist(text, name=None, default_name=None, default_city=None,
                 default_state=None, default_country=None, artist=None):
    if text:
        print(text)

    if name is None:
        name = prompt(Fg.green("Artist name"), default=default_name)

    origin = {}
    for o in ("city", "state", "country"):
        origin["origin_%s" % o] = prompt("   %s" % Fg.green(o.title()),
                                         default=locals()["default_%s" % o],
                                         required=False)

    if not artist:
        artist = Artist(name=name, **origin)
    else:
        artist.name = name
        for o in origin:
            setattr(artist, o, origin[o])
    return artist



from pathlib import Path
from configparser import ExtendedInterpolation
import nicfit
from .orm import MAIN_LIB_NAME

WEB_PORT = 6229
MAIN_SECT = "mishmash"
SA_KEY = "sqlalchemy.url"
CONFIG_ENV_VAR = "MISHMASH_CONFIG"
SQLITE_DB_URL = "sqlite:///{0}/mishmash.db".format(Path().cwd())
POSTGRES_DB_URL = "postgresql://mishmash@localhost/mishmash"
LOG_FORMAT = "<%(name)s:%(threadName)s> [%(levelname)s]: %(message)s"

LOGGING_CONFIG = (
    nicfit.logger.FileConfig(level="WARNING")
                 .addPackageLogger("alembic")
                 .addPackageLogger("mishmash")
                 .addPackageLogger("eyed3", pkg_level="ERROR")
                 .addHandler("file", "StreamHandler", args=("sys.stderr",))
)

DEFAULT_CONFIG = f"""
[mishmash]
sqlalchemy.url = {SQLITE_DB_URL}
;sqlalchemy.url = {POSTGRES_DB_URL}

# All sync'd media is assigned to the '{MAIN_LIB_NAME}' library unless
# instructed
;[library:{MAIN_LIB_NAME}]
;sync = true
;paths = dir1
;        dir2
;        dir_glob
# Directories to exclude, each as a regex
;excludes = dir_regex1
;           dir_regex2


[app:main]
use = call:mishmash.web:main
pyramid.reload_templates = true
pyramid.default_locale_name = en
pyramid.includes =
    pyramid_tm
# Devel opts
;pyramid.debug_authorization = false
;pyramid.debug_notfound = false
;pyramid.debug_routematch = false
;pyramid.includes =
;    pyramid_debugtoolbar
;    pyramid_tm

[server:main]
use = egg:waitress#main
host = 0.0.0.0
port = {WEB_PORT}

{LOGGING_CONFIG}
"""


class MusicLibrary:
    def __init__(self, name, paths=None, excludes=None, sync=True):
        self.name = name
        self.paths = paths or []
        self.sync = sync
        self.excludes = excludes

    @staticmethod
    def fromConfig(config):
        all_paths = []
        paths = config.get("paths")
        if paths:
            paths = paths.split("\n")
            for p in [Path(p).expanduser() for p in paths]:
                glob_paths = [
                    str(p) for p in Path("/").glob(str(p.relative_to("/")))
                ]
                all_paths += glob_paths if glob_paths else [str(p)]

        excludes = [str(Path(p).expanduser())
                        for p in config.getlist("excludes", fallback=[])]

        return MusicLibrary(config.name.split(":", 1)[1], paths=all_paths,
                            excludes=excludes,
                            sync=config.getboolean("sync", True))


class Config(nicfit.Config):
    def __init__(self, filename, **kwargs):
        super().__init__(filename, interpolation=ExtendedInterpolation(),
                         **kwargs)

    # XXX: new decorator could simplify these accessors.
    @property
    def db_url(self):
        return self.get(MAIN_SECT, SA_KEY).strip()

    @property
    def music_libs(self):
        for sect in [s for s in self.sections() if s.startswith("library:")]:
            yield MusicLibrary.fromConfig(self[sect])



import multiprocessing

from time import time
from pathlib import Path
from itertools import chain

from inotify.adapters import Inotify
from inotify.constants import (IN_ACCESS, IN_ALL_EVENTS, IN_ATTRIB,
                               IN_CLOSE_WRITE, IN_CLOSE_NOWRITE, IN_CREATE,
                               IN_DELETE, IN_ISDIR, IN_OPEN, IN_MODIFY,
                               IN_MOVED_TO, IN_MOVED_FROM)
from nicfit import getLogger

log = getLogger(__name__)
SYNC_INTERVAL = 10


class Monitor(multiprocessing.Process):

    def __init__(self):
        self._inotify = Inotify()
        self._inotify_mask = IN_ALL_EVENTS & (~IN_ACCESS &
                                              ~IN_OPEN &
                                              ~IN_CLOSE_NOWRITE &
                                              ~IN_CLOSE_WRITE)

        self._dir_queue = multiprocessing.Queue()
        self._sync_queue = multiprocessing.Queue()
        self._watched_dirs = {}   # {lib_name: set(dirs)}

        super().__init__(target=self._main, args=(self._dir_queue,
                                                  self._sync_queue))

    def _main(self, dir_queue, sync_queue):
        next_sync_t = time() + SYNC_INTERVAL
        sync_dirs = set()

        try:
            while True:
                # Check for new directories to watch
                while not dir_queue.empty():
                    lib, path = dir_queue.get()

                    watched = (path in set(chain(*self._watched_dirs.values())))

                    if lib not in self._watched_dirs:
                        self._watched_dirs[lib] = set()
                    self._watched_dirs[lib].add(path)

                    if not watched:
                        self._inotify.add_watch(str(path), self._inotify_mask)
                        log.info(f"Watching {path} (lib: {lib}) "
                                 f"(total dirs: {len(self._watched_dirs[lib])}")

                # Process Inotify
                for event in self._inotify.event_gen():
                    if event is None:
                        break

                    (header,
                     type_names,
                     watch_path,
                     filename) = event
                    watch_path = Path(watch_path)
                    filename = Path(filename)

                    log.debug(
                        f"WD=({header.wd}) MASK=({header.mask}) "
                        f"MASK->NAMES={type_names} WATCH-PATH={watch_path} FILENAME={filename}")

                    if header.mask & (IN_ATTRIB | IN_CREATE | IN_DELETE |
                                      IN_MODIFY | IN_MOVED_TO | IN_MOVED_FROM):
                        if IN_ISDIR & header.mask and header.mask & IN_CREATE:
                            watch_path = watch_path / filename
                        elif IN_ISDIR & header.mask and header.mask & IN_DELETE:
                            self._inotify.remove_watch(str(watch_path))

                        sync_dirs.add(watch_path)

                def _reqSync(l, d):
                    if d.exists():
                        sync_queue.put((lib, d))
                        log.info(f"Requesting sync {d} (lib: {l})")

                if time() > next_sync_t:
                    for d in sync_dirs:
                        for lib in self._watched_dirs:
                            lib_paths = self._watched_dirs[lib]
                            if d in lib_paths:
                                _reqSync(lib, d)
                                if not d.exists():
                                    self._watched_dirs[lib].remove(d)
                            elif d.parent in lib_paths:
                                _reqSync(lib, d)
                                self.dir_queue.put((lib, d))

                    sync_dirs.clear()
                    next_sync_t = time() + SYNC_INTERVAL

        except KeyboardInterrupt:
            pass
        finally:
            for path in set(chain(*self._watched_dirs.values())):
                self._inotify.remove_watch(str(path))

    @property
    def dir_queue(self):
        return self._dir_queue

    @property
    def sync_queue(self):
        return self._sync_queue



from eyed3.core import EP_MAX_SIZE_HINT
from nicfit import getLogger
from nicfit.command import Command as BaseCommand, CommandError

__all__ = ["Command", "CommandError", "EP_MAX_SIZE_HINT"]
log = getLogger(__name__)


class Command(BaseCommand):
    """Base class for MishMash commands."""
    _library_arg_nargs = None  # '*', '+', '?', None, 1

    config = None
    db_conn = None
    db_engine = None
    db_session = None

    def _initArgParser(self, parser):
        super()._initArgParser(parser)

        if self._library_arg_nargs:
            from .util import addLibraryArguments
            addLibraryArguments(parser, self._library_arg_nargs)

    def run(self, args):
        from . import database

        self.config = args.config

        (self.db_engine,
         SessionMaker,
         self.db_conn) = database.init(self.config.db_url)

        self.db_session = SessionMaker()

        try:
            retval = super().run(args)
            self.db_session.commit()
        except Exception:
            self.db_session.rollback()
            raise
        finally:
            self.db_session.close()
            self.db_conn.close()

        return retval



from collections import namedtuple


def __parse_version(v):                                       # pragma: nocover
    ver, rel = v, "final"
    for c in ("a", "b", "c"):
        parsed = v.split(c)
        if len(parsed) == 2:
            ver, rel = (parsed[0], c + parsed[1])

    v = tuple((int(v) for v in ver.split(".")))
    ver_info = namedtuple("Version", "major, minor, maint, release")(
        *(v + (tuple((0,)) * (3 - len(v))) + tuple((rel,))))
    return ver, rel, ver_info


__version__ = "0.3.1"
__release_name__ = ""
__years__ = "2013-2020"

_, __release__, __version_info__ = __parse_version(__version__)
__project_name__ = "MishMash"
__project_slug__ = "mishmash"
__pypi_name__ = "MishMash"
__author__ = "Travis Shirk"
__author_email__ = "travis@pobox.com"
__url__ = "https://github.com/nicfit/MishMash"
__description__ = "Music database and web interface."
__long_description__ = ""
__license__ = "GNU GPL v3.0"
__github_url__ = "https://github.com/nicfit/mishmash",
__version_txt__ = """
%(__name__)s %(__version__)s (C) Copyright %(__years__)s %(__author__)s
This program comes with ABSOLUTELY NO WARRANTY! See LICENSE for details.
Run with --help/-h for usage information or read the docs at
%(__url__)s
""" % (locals())



# -*- coding: utf-8 -*-
import mishmash
"""
test_mishmash
----------------------------------

Tests for `mishmash` module.
"""


def test_metadata():
    assert mishmash.version
    assert mishmash.__about__.__license__
    assert mishmash.__about__.__project_name__
    assert mishmash.__about__.__author__
    assert mishmash.__about__.__author_email__
    assert mishmash.__about__.__version__
    assert mishmash.__about__.__version_info__
    assert mishmash.__about__.__release__
    assert mishmash.__about__.__version_txt__


def test_database_fixture(database):
    assert database.url
    assert database.engine
    assert database.SessionMaker


def test_session_fixture(session):
    assert session



from pathlib import Path
from collections import namedtuple
import nicfit
import alembic
import alembic.config
from nicfit.util import cd
from sqlalchemy import create_engine, or_
from sqlalchemy.orm import sessionmaker, scoped_session
from sqlalchemy.orm.exc import NoResultFound

from sqlalchemy_utils.functions import (database_exists,
                                        create_database,
                                        drop_database)

from .orm import TYPES
from .orm import Artist, Track, Album, Tag
from .util import safeDbUrl

DEFAULT_ENGINE_ARGS = {"convert_unicode": True,
                       "encoding": "utf8",
                       "echo": False,
                      }
DEFAULT_SESSION_ARGS = {
        }

log = nicfit.getLogger(__name__)
DatabaseInfo = namedtuple("DatabaseInfo", ["engine",
                                           "SessionMaker",
                                           "connection"])


def init(db_url, engine_args=None, session_args=None, trans_mgr=None, scoped=False):
    alembic_d = Path(__file__).parent
    alembic_cfg = alembic.config.Config(str(alembic_d / "alembic.ini"))
    alembic_cfg.set_main_option("sqlalchemy.url", db_url)

    log.debug(f"Checking for database '{safeDbUrl(db_url)}'")
    if not database_exists(db_url):
        log.info(f"Creating database '{safeDbUrl(db_url)}'")
        create_database(db_url, template="template0")

    log.debug(f"Connecting to database '{safeDbUrl(db_url)}'")
    args = engine_args or DEFAULT_ENGINE_ARGS
    engine = create_engine(db_url, **args)
    connection = engine.connect()

    args = session_args or DEFAULT_SESSION_ARGS
    SessionMaker = sessionmaker(bind=engine, **args)
    if scoped:
        SessionMaker = scoped_session(SessionMaker)

    if trans_mgr:
        trans_mgr(SessionMaker)

    for T in TYPES:
        T.metadata.bind = engine

    # Upgrade to head (i.e. this) revision, or no-op if they match
    with cd(str(alembic_d)):
        alembic.command.upgrade(alembic_cfg, "head")

    return DatabaseInfo(engine, SessionMaker, connection)


def dropAll(url):
    drop_database(url)


"""''''## ###########################################################################3
## Works-in-progress, subject to change
## ###########################################################################3
"""


def getTag(t, session, lid, add=False):
    tag = None
    t = t[:Tag.NAME_LIMIT]
    try:
        tag = session.query(Tag).filter_by(name=t, lib_id=lid).one()
    except NoResultFound:
        if add:
            tag = Tag(name=t, lib_id=lid)
            session.add(tag)
            session.flush()
    return tag


def search(session, query):
    """Naive search of the database for `query`.

    :return: A dict with keys 'artists', 'albums', and 'tracks'. Each containing a list
             of the respective ORM type.
    """
    flat_query = "".join(query.split())

    artists = session.query(Artist).filter(
            or_(Artist.name.ilike(f"%%{query}%%"),
                Artist.name.ilike(f"%%{flat_query}%%"))
               ).all()
    albums = session.query(Album).filter(
            Album.title.ilike(f"%%{query}%%")).all()
    tracks = session.query(Track).filter(
            Track.title.ilike(f"%%{query}%%")).all()

    return dict(artists=artists,
                albums=albums,
                tracks=tracks)



#!/usr/bin/env python
import io
import os
import re
import sys
import warnings
from setuptools import setup, find_packages
from setuptools.command.install import install

classifiers = [
    "Intended Audience :: End Users/Desktop",
    "Operating System :: POSIX",
    "Natural Language :: English",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
]


def getPackageInfo():
    info_dict = {}
    info_keys = ["version", "name", "author", "author_email", "url", "license",
                 "description", "release_name", "github_url"]
    # FIXME: This remap is the exception, not the rule.
    key_remap = {"name": "pypi_name"}

    # __about__
    info_fpath = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                              ".",
                              "mishmash",
                              "__about__.py")
    with io.open(info_fpath, encoding='utf-8') as infof:
        for line in infof:
            for what in info_keys:
                rex = re.compile(r"__{what}__\s*=\s*['\"](.*?)['\"]"
                                  .format(what=what if what not in key_remap
                                                    else key_remap[what]))

                m = rex.match(line.strip())
                if not m:
                    continue
                info_dict[what] = m.groups()[0]

    vparts = info_dict["version"].split("-", maxsplit=1)
    info_dict["release"] = vparts[1] if len(vparts) > 1 else "final"

    # Requirements
    requirements, extras = requirements_yaml()
    info_dict["install_requires"] = requirements["main"] \
                                        if "main" in requirements else []
    info_dict["tests_require"] = requirements["test"] \
                                     if "test" in requirements else []
    info_dict["extras_require"] = extras

    # Info
    readme = ""
    if os.path.exists("README.rst"):
        with io.open("README.rst", encoding='utf-8') as readme_file:
            readme = readme_file.read()

    hist = "`changelog <https://github.com/nicfit/MishMash/blob/master/HISTORY.rst>`_"
    info_dict["long_description"] =\
        readme + "\n\n" +\
        "See the {} file for release history and changes.".format(hist)

    return info_dict, requirements


def requirements_yaml():
    prefix = "extra_"
    reqs = {}
    reqfile = os.path.join("requirements", "requirements.yml")
    if os.path.exists(reqfile):
        with io.open(reqfile, encoding='utf-8') as fp:
            curr = None
            for line in [l for l in [l.strip() for l in fp.readlines()]
                     if l and not l.startswith("#")]:
                if curr is None or line[0] != "-":
                    curr = line.split(":")[0]
                    reqs[curr] = []
                else:
                    assert line[0] == "-"
                    r = line[1:].strip()
                    if r:
                        reqs[curr].append(r.strip())

    return (reqs, {x[len(prefix):]: vals
                     for x, vals in reqs.items() if x.startswith(prefix)})


class PipInstallCommand(install, object):
    def run(self):
        reqs = " ".join(["'%s'" % r for r in PKG_INFO["install_requires"]])
        os.system("pip install " + reqs)
        # XXX: py27 compatible
        return super(PipInstallCommand, self).run()


PKG_INFO, REQUIREMENTS = getPackageInfo()
if PKG_INFO["release"].startswith("a"):
    #classifiers.append("Development Status :: 1 - Planning")
    #classifiers.append("Development Status :: 2 - Pre-Alpha")
    classifiers.append("Development Status :: 3 - Alpha")
elif PKG_INFO["release"].startswith("b"):
    classifiers.append("Development Status :: 4 - Beta")
else:
    classifiers.append("Development Status :: 5 - Production/Stable")
    #classifiers.append("Development Status :: 6 - Mature")
    #classifiers.append("Development Status :: 7 - Inactive")

gz = "{name}-{version}.tar.gz".format(**PKG_INFO)
PKG_INFO["download_url"] = (
    "{github_url}/releases/downloads/v{version}/{gz}"
    .format(gz=gz, **PKG_INFO)
)


def package_files(directory, prefix=".."):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        if "__pycache__" in path:
            continue
        for filename in filenames:
            if filename.endswith(".pyc"):
                continue
            paths.append(os.path.join(prefix, path, filename))
    return paths


if sys.argv[1:] and sys.argv[1] == "--release-name":
    print(PKG_INFO["release_name"])
    sys.exit(0)
else:
    # The extra command line options we added cause warnings, quell that.
    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", message="Unknown distribution option")
        warnings.filterwarnings("ignore", message="Normalizing")
        setup(classifiers=classifiers,
              package_dir={"": "."},
              packages=find_packages(".",
                                     exclude=["tests", "tests.*"]),
              zip_safe=False,
              platforms=["Any"],
              keywords=["music", "database"],
              test_suite="./tests",
              include_package_data=True,
              package_data={
                  "mishmash": ["alembic.ini"] +
                              package_files("mishmash/alembic"),
                  "mishmash.web": package_files("mishmash/web/static",
                                                "../..") +
                                  package_files("mishmash/web/templates",
                                                "../.."),
              },
              entry_points={
                  "console_scripts": [
                      "mishmash = mishmash.__main__:app.run",
                  ]
              },
              cmdclass={
                  "install": PipInstallCommand,
              },
              **PKG_INFO
        )



import os
import nicfit
from nicfit.console import pout
from nicfit.console.ansi import Fg

from ...orm import VARIOUS_ARTISTS_ID
from ...orm import Artist, Track, Album

log = nicfit.getLogger(__name__)


def deleteOrphans(session):
    num_orphaned_artists = 0
    num_orphaned_albums = 0
    num_orphaned_tracks = 0
    found_ids = set()

    # Tracks
    for track in session.query(Track).all():
        if not os.path.exists(track.path):
            pout(Fg.red("Removing track") + ": " + track.path)
            session.delete(track)
            num_orphaned_tracks += 1
            log.warn("Deleting track: %s" % str(track))
    session.flush()

    # Albums
    found_ids.clear()
    for album in session.query(Album).all():
        if album.id in found_ids:
            continue

        any_track = session.query(Track).filter(Track.album_id == album.id).first()
        if not any_track:
            log.warn("Deleting album: %s" % str(album))
            session.delete(album)
            num_orphaned_albums += 1
        else:
            found_ids.add(album.id)
    session.flush()

    # Artists
    found_ids.clear()
    for artist in session.query(Artist).all():
        if (artist.id == VARIOUS_ARTISTS_ID or
                artist.id in found_ids):
            continue

        any_track = session.query(Track).filter(Track.artist_id == artist.id) \
                                        .first()
        any_album = session.query(Album).filter(Album.artist_id == artist.id) \
                                        .first()
        if not any_track and (not any_album or not any_album.tracks):
            log.warn("Deleting artist: %s" % str(artist))
            session.delete(artist)
            num_orphaned_artists += 1
        else:
            found_ids.add(artist.id)
    session.flush()

    return (num_orphaned_tracks, num_orphaned_artists, num_orphaned_albums)


def syncImage(img, current, session):
    """Add or updated the Image."""
    def _img_str(i):
        return "%s - %s" % (i.type, i.description)

    for db_img in current.images:
        img_info = (img.type, img.md5, img.size)
        db_img_info = (db_img.type, db_img.md5, db_img.size)

        if db_img_info == img_info:
            img = None
            break
        elif (db_img.type == img.type and
                db_img.description == img.description):

            if img.md5 != db_img.md5:
                # Update image
                current.images.remove(db_img)
                current.images.append(img)
                session.add(current)
                pout(Fg.green("Updating image") + ": " + _img_str(img))
            img = None
            break

    if img:
        # Add image
        current.images.append(img)
        session.add(current)
        pout(Fg.green("Adding image") + ": " + _img_str(img))
